@@ -46,6 +46,11 @@ class BinaryArch(NamedTuple):
     isa: ISA = ISA.AMD64
 
 
+class CompilerType(Enum):
+    gcc = 'gcc'
+    clang = 'clang'
+    unknown = 'unknown'
+
 
 class Env:
 
@@ -81,6 +86,7 @@ class Env:
         self.binary_arch = binary_arch
         self.native_optimizations = native_optimizations
         self._cc_version_string = ''
+        self._compiler_type: Optional[CompilerType] = None
 
     @property
     def cc_version_string(self) -> str:
@@ -89,8 +95,16 @@ class Env:
         return self._cc_version_string
 
     @property
-    def is_gcc(self) -> bool:
-        return 'gcc' in self.cc_version_string.split(maxsplit=1)[0].lower()
+    def compiler_type(self) -> CompilerType:
+        if self._compiler_type is None:
+            raw = self.cc_version_string
+            if 'Free Software Foundation' in raw:
+                self._compiler_type = CompilerType.gcc
+            elif 'clang' in raw.lower().split():
+                self._compiler_type = CompilerType.clang
+            else:
+                self._compiler_type = CompilerType.unknown
+        return self._compiler_type
 
     def copy(self) -> 'Env':
         ans = Env(self.cc, list(self.cppflags), list(self.cflags), list(self.ldflags), dict(self.library_paths), list(self.ldpaths), self.ccver)
@@ -286,7 +300,7 @@ def generate_wrappers(glfw_header: str) -> None:
     int glfwCocoaSetBackgroundBlur(GLFWwindow *w, int blur_radius)
     bool glfwSetX11WindowBlurred(GLFWwindow *w, bool enable_blur)
     void* glfwGetX11Display(void)
-    int32_t glfwGetX11Window(GLFWwindow* window)
+    unsigned long glfwGetX11Window(GLFWwindow* window)
     void glfwSetPrimarySelectionString(GLFWwindow* window, const char* string)
     void glfwCocoaSetWindowChrome(GLFWwindow* window, unsigned int color, bool use_system_color, unsigned int system_color,\
     int background_blur, unsigned int hide_window_decorations, bool show_text_in_titlebar, int color_space, float background_opacity, bool resizable)
