@@ -30,7 +30,6 @@
 
 #include "internal.h"
 #include "backend_utils.h"
-#include "memfd.h"
 #include "linux_notify.h"
 #include "wl_client_side_decorations.h"
 #include "../kitty/monotonic.h"
@@ -152,14 +151,63 @@ static struct wl_buffer* createShmBuffer(const GLFWimage* image, bool is_opaque,
     return buffer;
 }
 
+static int
+glfw_cursor_shape_to_wayland_cursor_shape(GLFWCursorShape g) {
+#define C(g, w) case g: return w;
+    switch(g) {
+        /* start glfw to wayland mapping (auto generated by gen-key-constants.py do not edit) */
+        C(GLFW_DEFAULT_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_DEFAULT);
+        C(GLFW_TEXT_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_TEXT);
+        C(GLFW_POINTER_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_POINTER);
+        C(GLFW_HELP_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_HELP);
+        C(GLFW_WAIT_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_WAIT);
+        C(GLFW_PROGRESS_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_PROGRESS);
+        C(GLFW_CROSSHAIR_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_CROSSHAIR);
+        C(GLFW_CELL_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_CELL);
+        C(GLFW_VERTICAL_TEXT_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_VERTICAL_TEXT);
+        C(GLFW_MOVE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_MOVE);
+        C(GLFW_E_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_E_RESIZE);
+        C(GLFW_NE_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_NE_RESIZE);
+        C(GLFW_NW_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_NW_RESIZE);
+        C(GLFW_N_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_N_RESIZE);
+        C(GLFW_SE_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_SE_RESIZE);
+        C(GLFW_SW_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_SW_RESIZE);
+        C(GLFW_S_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_S_RESIZE);
+        C(GLFW_W_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_W_RESIZE);
+        C(GLFW_EW_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_EW_RESIZE);
+        C(GLFW_NS_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_NS_RESIZE);
+        C(GLFW_NESW_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_NESW_RESIZE);
+        C(GLFW_NWSE_RESIZE_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_NWSE_RESIZE);
+        C(GLFW_ZOOM_IN_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_ZOOM_IN);
+        C(GLFW_ZOOM_OUT_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_ZOOM_OUT);
+        C(GLFW_ALIAS_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_ALIAS);
+        C(GLFW_COPY_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_COPY);
+        C(GLFW_NOT_ALLOWED_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_NOT_ALLOWED);
+        C(GLFW_NO_DROP_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_NO_DROP);
+        C(GLFW_GRAB_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_GRAB);
+        C(GLFW_GRABBING_CURSOR, WP_CURSOR_SHAPE_DEVICE_V1_SHAPE_GRABBING);
+/* end glfw to wayland mapping */
+        default: return -1;
+    }
+#undef C
+}
+
 static void
 setCursorImage(_GLFWwindow* window, bool on_theme_change) {
-    _GLFWcursorWayland defaultCursor = {.shape = GLFW_ARROW_CURSOR};
+    _GLFWcursorWayland defaultCursor = {.shape = GLFW_DEFAULT_CURSOR};
     _GLFWcursorWayland* cursorWayland = window->cursor ? &window->cursor->wl : &defaultCursor;
+    if (_glfw.wl.wp_cursor_shape_device_v1) {
+        int which = glfw_cursor_shape_to_wayland_cursor_shape(cursorWayland->shape);
+        if (which > -1) {
+            wp_cursor_shape_device_v1_set_shape(_glfw.wl.wp_cursor_shape_device_v1, _glfw.wl.pointer_enter_serial, (uint32_t)which);
+            return;
+        }
+    }
     struct wl_cursor_image* image = NULL;
     struct wl_buffer* buffer = NULL;
     struct wl_surface* surface = _glfw.wl.cursorSurface;
     const int scale = window->wl.scale;
+    if (!_glfw.wl.pointer) return;
 
     if (cursorWayland->scale < 0) {
         buffer = cursorWayland->buffer;
@@ -717,7 +765,7 @@ static void incrementCursorImage(_GLFWwindow* window)
 {
     if (window && window->wl.decorations.focus == CENTRAL_WINDOW && window->cursorMode != GLFW_CURSOR_HIDDEN) {
         _GLFWcursor* cursor = window->wl.currentCursor;
-        if (cursor && cursor->wl.cursor)
+        if (cursor && cursor->wl.cursor && cursor->wl.cursor->image_count)
         {
             cursor->wl.currentImage += 1;
             cursor->wl.currentImage %= cursor->wl.cursor->image_count;
@@ -831,16 +879,38 @@ struct wl_cursor* _glfwLoadCursor(GLFWCursorShape shape, struct wl_cursor_theme*
     struct wl_cursor* ans = NULL;
     switch (shape)
     {
-        C(GLFW_ARROW_CURSOR, "left_ptr", "arrow", "default")
-        C(GLFW_IBEAM_CURSOR, "xterm", "ibeam", "text")
-        C(GLFW_CROSSHAIR_CURSOR, "crosshair", "cross")
-        C(GLFW_HAND_CURSOR, "hand2", "grab", "grabbing", "closedhand")
-        C(GLFW_HRESIZE_CURSOR, "sb_h_double_arrow", "h_double_arrow", "col-resize")
-        C(GLFW_VRESIZE_CURSOR, "sb_v_double_arrow", "v_double_arrow", "row-resize")
-        C(GLFW_NW_RESIZE_CURSOR, "top_left_corner", "nw-resize")
-        C(GLFW_NE_RESIZE_CURSOR, "top_right_corner", "ne-resize")
-        C(GLFW_SW_RESIZE_CURSOR, "bottom_left_corner", "sw-resize")
-        C(GLFW_SE_RESIZE_CURSOR, "bottom_right_corner", "se-resize")
+        /* start glfw to xc mapping (auto generated by gen-key-constants.py do not edit) */
+        C(GLFW_DEFAULT_CURSOR, "default", "left_ptr");
+        C(GLFW_TEXT_CURSOR, "text", "xterm", "ibeam");
+        C(GLFW_POINTER_CURSOR, "pointing_hand", "pointer", "hand2", "hand");
+        C(GLFW_HELP_CURSOR, "help", "question_arrow", "whats_this");
+        C(GLFW_WAIT_CURSOR, "wait", "clock", "watch");
+        C(GLFW_PROGRESS_CURSOR, "progress", "half-busy", "left_ptr_watch");
+        C(GLFW_CROSSHAIR_CURSOR, "crosshair", "tcross");
+        C(GLFW_CELL_CURSOR, "cell", "plus", "cross");
+        C(GLFW_VERTICAL_TEXT_CURSOR, "vertical-text");
+        C(GLFW_MOVE_CURSOR, "move", "fleur", "pointer-move");
+        C(GLFW_E_RESIZE_CURSOR, "e-resize", "right_side");
+        C(GLFW_NE_RESIZE_CURSOR, "ne-resize", "top_right_corner");
+        C(GLFW_NW_RESIZE_CURSOR, "nw-resize", "top_left_corner");
+        C(GLFW_N_RESIZE_CURSOR, "n-resize", "top_side");
+        C(GLFW_SE_RESIZE_CURSOR, "se-resize", "bottom_right_corner");
+        C(GLFW_SW_RESIZE_CURSOR, "sw-resize", "bottom_left_corner");
+        C(GLFW_S_RESIZE_CURSOR, "s-resize", "bottom_side");
+        C(GLFW_W_RESIZE_CURSOR, "w-resize", "left_side");
+        C(GLFW_EW_RESIZE_CURSOR, "ew-resize", "sb_h_double_arrow", "split_h");
+        C(GLFW_NS_RESIZE_CURSOR, "ns-resize", "sb_v_double_arrow", "split_v");
+        C(GLFW_NESW_RESIZE_CURSOR, "nesw-resize", "size_bdiag", "size-bdiag");
+        C(GLFW_NWSE_RESIZE_CURSOR, "nwse-resize", "size_fdiag", "size-fdiag");
+        C(GLFW_ZOOM_IN_CURSOR, "zoom-in", "zoom_in");
+        C(GLFW_ZOOM_OUT_CURSOR, "zoom-out", "zoom_out");
+        C(GLFW_ALIAS_CURSOR, "dnd-link");
+        C(GLFW_COPY_CURSOR, "dnd-copy");
+        C(GLFW_NOT_ALLOWED_CURSOR, "not-allowed", "forbidden", "crossed_circle");
+        C(GLFW_NO_DROP_CURSOR, "no-drop", "dnd-no-drop");
+        C(GLFW_GRAB_CURSOR, "grab", "openhand", "hand1");
+        C(GLFW_GRABBING_CURSOR, "grabbing", "closedhand", "dnd-none");
+/* end glfw to xc mapping */
         case GLFW_INVALID_CURSOR:
             break;
     }
@@ -1920,22 +1990,6 @@ static const struct zwp_primary_selection_device_v1_listener primary_selection_d
 };
 
 
-static void
-clipboard_copy_callback_done(void *data, struct wl_callback *callback, uint32_t serial) {
-    if (_glfw.wl.dataDevice && data == (void*)_glfw.wl.dataSourceForClipboard) {
-        wl_data_device_set_selection(_glfw.wl.dataDevice, data, serial);
-    }
-    wl_callback_destroy(callback);
-}
-
-static void
-primary_selection_copy_callback_done(void *data, struct wl_callback *callback, uint32_t serial) {
-    if (_glfw.wl.primarySelectionDevice && data == (void*)_glfw.wl.dataSourceForPrimarySelection) {
-        zwp_primary_selection_device_v1_set_selection(_glfw.wl.primarySelectionDevice, data, serial);
-    }
-    wl_callback_destroy(callback);
-}
-
 void _glfwSetupWaylandDataDevice(void) {
     _glfw.wl.dataDevice = wl_data_device_manager_get_data_device(_glfw.wl.dataDeviceManager, _glfw.wl.seat);
     if (_glfw.wl.dataDevice) wl_data_device_add_listener(_glfw.wl.dataDevice, &data_device_listener, NULL);
@@ -2027,13 +2081,25 @@ _glfwPlatformSetClipboard(GLFWClipboardType t) {
         }
         f(data_source, cd->mime_types[i]);
     }
-    struct wl_callback *callback = wl_display_sync(_glfw.wl.display);
     if (t == GLFW_CLIPBOARD) {
-        static const struct wl_callback_listener clipboard_copy_callback_listener = {.done = clipboard_copy_callback_done};
-        wl_callback_add_listener(callback, &clipboard_copy_callback_listener, _glfw.wl.dataSourceForClipboard);
+        // According to some interpretations of the Wayland spec only the application that has keyboard focus can set the clipboard.
+        // Hurray for the Wayland nanny state!
+        //
+        // However in wl-roots based compositors, using the serial from the keyboard enter event doesn't work. No clue what
+        // the correct serial to use here is. Given this Wayland there probably isn't one. What a joke.
+        // Bug report: https://github.com/kovidgoyal/kitty/issues/6890
+        // Ironically one of the contributors to wl_roots claims the keyboard enter serial is the correct one to use:
+        // https://emersion.fr/blog/2020/wayland-clipboard-drag-and-drop/
+        // The Wayland spec itself says "serial number of the event that triggered this request"
+        // https://wayland.freedesktop.org/docs/html/apa.html#protocol-spec-wl_data_device
+        // So who the fuck knows. Just use the latest received serial and ask anybody that uses Wayland
+        // to get their head examined.
+        wl_data_device_set_selection(_glfw.wl.dataDevice, _glfw.wl.dataSourceForClipboard, _glfw.wl.serial);
     } else {
-        static const struct wl_callback_listener primary_selection_copy_callback_listener = {.done = primary_selection_copy_callback_done};
-        wl_callback_add_listener(callback, &primary_selection_copy_callback_listener, _glfw.wl.dataSourceForPrimarySelection);
+        // According to the Wayland spec we can only set the primary selection in response to a pointer button event
+        // Hurray for the Wayland nanny state!
+        zwp_primary_selection_device_v1_set_selection(
+                _glfw.wl.primarySelectionDevice, _glfw.wl.dataSourceForPrimarySelection, _glfw.wl.pointer_serial);
     }
 }
 
@@ -2266,3 +2332,8 @@ GLFWAPI bool glfwWaylandSetTitlebarColor(GLFWwindow *handle, uint32_t color, boo
     }
     return false;
 }
+
+GLFWAPI void glfwWaylandRedrawCSDWindowTitle(GLFWwindow *handle) {
+    _GLFWwindow* window = (_GLFWwindow*) handle;
+    change_csd_title(window);
+}
