@@ -110,9 +110,9 @@ init_rsync(Rsync *ans, size_t block_size, int strong_hash_type, int checksum_typ
     ans->hasher = ans->hasher_constructor();
     ans->checksummer = ans->checksummer_constructor();
     ans->hasher.state = ans->hasher.new();
-    if (ans->hasher.state == NULL) { free(ans); return "Out of memory"; }
+    if (ans->hasher.state == NULL) { free_rsync(ans); return "Out of memory"; }
     ans->checksummer.state = ans->checksummer.new();
-    if (ans->checksummer.state == NULL) { free(ans); return "Out of memory"; }
+    if (ans->checksummer.state == NULL) { free_rsync(ans); return "Out of memory"; }
     return NULL;
 }
 
@@ -901,13 +901,6 @@ PyTypeObject Hasher_Type = {
 };
 // }}} end Hasher
 
-static PyObject*
-decode_utf8_buffer(PyObject *self UNUSED, PyObject *args) {
-    RAII_PY_BUFFER(buf);
-    if (!PyArg_ParseTuple(args, "s*", &buf)) return NULL;
-    return PyUnicode_FromStringAndSize(buf.buf, buf.len);
-}
-
 static bool
 call_ftc_callback(PyObject *callback, char *src, Py_ssize_t key_start, Py_ssize_t key_length, Py_ssize_t val_start, Py_ssize_t val_length) {
     while(src[key_start] == ';' && key_length > 0 ) { key_start++; key_length--; }
@@ -952,7 +945,6 @@ parse_ftc(PyObject *self UNUSED, PyObject *args) {
 
 static PyMethodDef module_methods[] = {
     {"parse_ftc", parse_ftc, METH_VARARGS, ""},
-    {"decode_utf8_buffer", decode_utf8_buffer, METH_VARARGS, ""},
     {NULL, NULL, 0, NULL}        /* Sentinel */
 };
 
