@@ -184,38 +184,79 @@ glyph_id_for_codepoint_ctfont(CTFontRef ct_font, char_type ch) {
     return glyphs[0];
 }
 
+static bool
+cf_string_equals(CFStringRef a, CFStringRef b) { return CFStringCompare(a, b, 0) == kCFCompareEqualTo; }
+
+#define LAST_RESORT_FONT_NAME "LastResort"
+
 static bool
 is_last_resort_font(CTFontRef new_font) {
     CFStringRef name = CTFontCopyPostScriptName(new_font);
-    CFComparisonResult cr = CFStringCompare(name, CFSTR("LastResort"), 0);
+    bool ans = cf_string_equals(name, CFSTR(LAST_RESORT_FONT_NAME));
     CFRelease(name);
-    return cr == kCFCompareEqualTo;
+    return ans;
+}
+
+static CTFontDescriptorRef _nerd_font_descriptor = NULL;
+
+static CTFontRef nerd_font(CGFloat sz) {
+    static bool searched = false;
+    if (!searched) {
+        searched = true;
+        CFArrayRef fonts = CTFontCollectionCreateMatchingFontDescriptors(all_fonts_collection());
+        const CFIndex count = CFArrayGetCount(fonts);
+        for (CFIndex i = 0; i < count; i++) {
+            CTFontDescriptorRef descriptor = (CTFontDescriptorRef)CFArrayGetValueAtIndex(fonts, i);
+            CFStringRef name = CTFontDescriptorCopyAttribute(descriptor, kCTFontNameAttribute);
+            bool is_nerd_font = cf_string_equals(name, CFSTR("SymbolsNFM"));
+            CFRelease(name);
+            if (is_nerd_font) {
+                _nerd_font_descriptor = CTFontDescriptorCreateCopyWithAttributes(descriptor, CTFontDescriptorCopyAttributes(descriptor));
+                break;
+            }
+        }
+        CFRelease(fonts);
+    }
+    return _nerd_font_descriptor ? CTFontCreateWithFontDescriptor(_nerd_font_descriptor, sz, NULL) : NULL;
+}
+
+static bool
+font_can_render_cell(CTFontRef font, CPUCell *cell) {
+    char_type ch = cell->ch ? cell->ch : ' ';
+    bool found = true;
+    if (!glyph_id_for_codepoint_ctfont(font, ch)) found = false;
+    for (unsigned i = 0; i < arraysz(cell->cc_idx) && cell->cc_idx[i] && found; i++) {
+        char_type cch = codepoint_for_mark(cell->cc_idx[i]);
+        if (!glyph_id_for_codepoint_ctfont(font, cch)) found = false;
+    }
+    return found;
 }
 
 static CTFontRef
 manually_search_fallback_fonts(CTFontRef current_font, CPUCell *cell) {
+    char_type ch = cell->ch ? cell->ch : ' ';
+    const bool in_first_pua = 0xe000 <= ch && ch <= 0xf8ff;
+    // preferentially load from NERD fonts
+    if (in_first_pua) {
+        CTFontRef nf = nerd_font(CTFontGetSize(current_font));
+        if (nf) {
+            if (font_can_render_cell(nf, cell)) return nf;
+            CFRelease(nf);
+        }
+    }
     CFArrayRef fonts = CTFontCollectionCreateMatchingFontDescriptors(all_fonts_collection());
     CTFontRef ans = NULL;
     const CFIndex count = CFArrayGetCount(fonts);
     for (CFIndex i = 0; i < count; i++) {
         CTFontDescriptorRef descriptor = (CTFontDescriptorRef)CFArrayGetValueAtIndex(fonts, i);
         CTFontRef new_font = CTFontCreateWithFontDescriptor(descriptor, CTFontGetSize(current_font), NULL);
-        if (new_font) {
-            if (!is_last_resort_font(new_font)) {
-                char_type ch = cell->ch ? cell->ch : ' ';
-                bool found = true;
-                if (!glyph_id_for_codepoint_ctfont(new_font, ch)) found = false;
-                for (unsigned i = 0; i < arraysz(cell->cc_idx) && cell->cc_idx[i] && found; i++) {
-                    ch = codepoint_for_mark(cell->cc_idx[i]);
-                    if (!glyph_id_for_codepoint_ctfont(new_font, ch)) found = false;
-                }
-                if (found) {
-                    ans = new_font;
-                    break;
-                }
+        if (!is_last_resort_font(new_font)) {
+            if (font_can_render_cell(new_font, cell)) {
+                ans = new_font;
+                break;
             }
-            CFRelease(new_font);
         }
+        CFRelease(new_font);
     }
     CFRelease(fonts);
     return ans;
@@ -231,16 +272,15 @@ find_substitute_face(CFStringRef str, CTFontRef old_font, CPUCell *cpu_cell) {
         CTFontRef new_font = CTFontCreateForString(old_font, str, CFRangeMake(start, amt));
         if (amt == len && len != 1) amt = 1;
         else start++;
-        if (new_font == NULL) { PyErr_SetString(PyExc_ValueError, "Failed to find fallback CTFont"); return NULL; }
         if (new_font == old_font) { CFRelease(new_font); continue; }
-        if (is_last_resort_font(new_font)) {
-            CFRelease(new_font);
+        if (!new_font || is_last_resort_font(new_font)) {
+            if (new_font) CFRelease(new_font);
             if (is_private_use(cpu_cell->ch)) {
                 // CoreTexts fallback font mechanism does not work for private use characters
                 new_font = manually_search_fallback_fonts(old_font, cpu_cell);
                 if (new_font) return new_font;
             }
-            PyErr_SetString(PyExc_ValueError, "Failed to find fallback CTFont other than the LastResort font");
+            PyErr_Format(PyExc_ValueError, "Failed to find fallback CTFont other than the %s font for: %s", LAST_RESORT_FONT_NAME, [(NSString *)str UTF8String]);
             return NULL;
         }
         return new_font;
@@ -249,8 +289,26 @@ find_substitute_face(CFStringRef str, CTFontRef old_font, CPUCell *cpu_cell) {
     return NULL;
 }
 
+static CTFontRef
+apply_styles_to_fallback_font(CTFontRef original_fallback_font, bool bold, bool italic) {
+    if (!original_fallback_font || is_last_resort_font(original_fallback_font)) return original_fallback_font;
+    CTFontRef ans = nil;
+    CTFontDescriptorRef original_descriptor = CTFontCopyFontDescriptor(original_fallback_font);
+    CTFontSymbolicTraits traits = kCTFontTraitMonoSpace;
+    if (bold) traits |= kCTFontTraitBold;
+    if (italic) traits |= kCTFontTraitItalic;
+    CTFontDescriptorRef descriptor = CTFontDescriptorCreateCopyWithSymbolicTraits(original_descriptor, traits, traits);
+    CFRelease(original_descriptor);
+    if (descriptor) {
+        ans = CTFontCreateWithFontDescriptor(descriptor, CTFontGetSize(original_fallback_font), NULL);
+        CFRelease(descriptor);
+    }
+    if (ans) { CFRelease(original_fallback_font); return ans; }
+    return original_fallback_font;
+}
+
 PyObject*
-create_fallback_face(PyObject *base_face, CPUCell* cell, bool UNUSED bold, bool UNUSED italic, bool emoji_presentation, FONTS_DATA_HANDLE fg) {
+create_fallback_face(PyObject *base_face, CPUCell* cell, bool bold, bool italic, bool emoji_presentation, FONTS_DATA_HANDLE fg) {
     CTFace *self = (CTFace*)base_face;
     CTFontRef new_font;
 #define search_for_fallback() \
@@ -268,9 +326,22 @@ create_fallback_face(PyObject *base_face, CPUCell* cell, bool UNUSED bold, bool
             search_for_fallback();
         }
     }
-    else { search_for_fallback(); }
+    else { search_for_fallback(); new_font = apply_styles_to_fallback_font(new_font, bold, italic); }
     if (new_font == NULL) return NULL;
-    return (PyObject*)ct_face(new_font, fg);
+    PyObject *postscript_name = Py_BuildValue("s", convert_cfstring(CTFontCopyPostScriptName(new_font), true));
+    ssize_t idx = -1;
+    PyObject *q, *ans = NULL;
+    while ((q = iter_fallback_faces(fg, &idx))) {
+        CTFace *qf = (CTFace*)q;
+        if (PyObject_RichCompareBool(postscript_name, qf->postscript_name, Py_EQ) == 1) {
+            ans = PyLong_FromSsize_t(idx);
+            break;
+        }
+    }
+    Py_CLEAR(postscript_name);
+    if (ans == NULL) return (PyObject*)ct_face(new_font, fg);
+    CFRelease(new_font);
+    return ans;
 }
 
 unsigned int
@@ -319,6 +390,9 @@ harfbuzz_font_for_face(PyObject* s) {
     if (!self->hb_font) {
         self->hb_font = hb_coretext_font_create(self->ct_font);
         if (!self->hb_font) fatal("Failed to create hb_font");
+        // dunno if we need this, harfbuzz docs say it is used by CoreText
+        // for optical sizing which changes the look of glyphs at small and large sizes
+        hb_font_set_ptem(self->hb_font, self->scaled_point_sz);
         hb_ot_font_set_funcs(self->hb_font);
     }
     return self->hb_font;
@@ -421,17 +495,17 @@ struct RenderBuffers {
     CGGlyph *glyphs;
     CGRect *boxes;
     CGPoint *positions;
-    CGSize *advances;
 };
 static struct RenderBuffers buffers = {0};
 
 static void
 finalize(void) {
-    free(buffers.render_buf); free(buffers.glyphs); free(buffers.boxes); free(buffers.positions); free(buffers.advances);
+    free(buffers.render_buf); free(buffers.glyphs); free(buffers.boxes); free(buffers.positions);
     memset(&buffers, 0, sizeof(struct RenderBuffers));
     if (all_fonts_collection_data) CFRelease(all_fonts_collection_data);
     if (window_title_font) CFRelease(window_title_font);
     window_title_font = nil;
+    if (_nerd_font_descriptor) CFRelease(_nerd_font_descriptor);
 }
 
 
@@ -471,11 +545,10 @@ ensure_render_space(size_t width, size_t height, size_t num_glyphs) {
     }
     if (buffers.sz < num_glyphs) {
         buffers.sz = MAX(128, num_glyphs * 2);
-        buffers.advances = calloc(sizeof(buffers.advances[0]), buffers.sz);
         buffers.boxes = calloc(sizeof(buffers.boxes[0]), buffers.sz);
         buffers.glyphs = calloc(sizeof(buffers.glyphs[0]), buffers.sz);
         buffers.positions = calloc(sizeof(buffers.positions[0]), buffers.sz);
-        if (!buffers.advances || !buffers.boxes || !buffers.glyphs || !buffers.positions) fatal("Out of memory");
+        if (!buffers.boxes || !buffers.glyphs || !buffers.positions) fatal("Out of memory");
     }
 }
 
@@ -612,7 +685,7 @@ render_single_ascii_char_as_mask(const char ch, size_t *result_width, size_t *re
 
 
 static bool
-do_render(CTFontRef ct_font, bool bold, bool italic, hb_glyph_info_t *info, hb_glyph_position_t *hb_positions, unsigned int num_glyphs, pixel *canvas, unsigned int cell_width, unsigned int cell_height, unsigned int num_cells, unsigned int baseline, bool *was_colored, bool allow_resize, FONTS_DATA_HANDLE fg, bool center_glyph) {
+do_render(CTFontRef ct_font, unsigned int units_per_em, bool bold, bool italic, hb_glyph_info_t *info, hb_glyph_position_t *hb_positions, unsigned int num_glyphs, pixel *canvas, unsigned int cell_width, unsigned int cell_height, unsigned int num_cells, unsigned int baseline, bool *was_colored, bool allow_resize, FONTS_DATA_HANDLE fg, bool center_glyph) {
     unsigned int canvas_width = cell_width * num_cells;
     ensure_render_space(canvas_width, cell_height, num_glyphs);
     CGRect br = CTFontGetBoundingRectsForGlyphs(ct_font, kCTFontOrientationHorizontal, buffers.glyphs, buffers.boxes, num_glyphs);
@@ -626,17 +699,20 @@ do_render(CTFontRef ct_font, bool bold, bool italic, hb_glyph_info_t *info, hb_g
             CGFloat sz = CTFontGetSize(ct_font);
             sz *= canvas_width / right;
             CTFontRef new_font = CTFontCreateCopyWithAttributes(ct_font, sz, NULL, NULL);
-            bool ret = do_render(new_font, bold, italic, info, hb_positions, num_glyphs, canvas, cell_width, cell_height, num_cells, baseline, was_colored, false, fg, center_glyph);
+            bool ret = do_render(new_font, CTFontGetUnitsPerEm(new_font), bold, italic, info, hb_positions, num_glyphs, canvas, cell_width, cell_height, num_cells, baseline, was_colored, false, fg, center_glyph);
             CFRelease(new_font);
             return ret;
         }
     }
     CGFloat x = 0, y = 0;
-    CTFontGetAdvancesForGlyphs(ct_font, kCTFontOrientationDefault, buffers.glyphs, buffers.advances, num_glyphs);
+    CGFloat scale = CTFontGetSize(ct_font) / units_per_em;
     for (unsigned i=0; i < num_glyphs; i++) {
-        buffers.positions[i].x = x; buffers.positions[i].y = y;
-        if (debug_rendering) printf("x=%f origin=%f width=%f advance=%f\n", x, buffers.boxes[i].origin.x, buffers.boxes[i].size.width, buffers.advances[i].width);
-        x += buffers.advances[i].width; y += buffers.advances[i].height;
+        buffers.positions[i].x = x + hb_positions[i].x_offset * scale; buffers.positions[i].y = y + hb_positions[i].y_offset * scale;
+        if (debug_rendering) printf("x=%f y=%f origin=%f width=%f x_advance=%f x_offset=%f y_advance=%f y_offset=%f\n",
+                buffers.positions[i].x, buffers.positions[i].y, buffers.boxes[i].origin.x, buffers.boxes[i].size.width,
+                hb_positions[i].x_advance * scale, hb_positions[i].x_offset * scale,
+                hb_positions[i].y_advance * scale, hb_positions[i].y_offset * scale);
+        x += hb_positions[i].x_advance * scale; y += hb_positions[i].y_advance * scale;
     }
     if (*was_colored) {
         render_color_glyph(ct_font, (uint8_t*)canvas, info[0].codepoint, cell_width * num_cells, cell_height, baseline);
@@ -661,7 +737,7 @@ render_glyphs_in_cells(PyObject *s, bool bold, bool italic, hb_glyph_info_t *inf
     CTFace *self = (CTFace*)s;
     ensure_render_space(128, 128, num_glyphs);
     for (unsigned i=0; i < num_glyphs; i++) buffers.glyphs[i] = info[i].codepoint;
-    return do_render(self->ct_font, bold, italic, info, hb_positions, num_glyphs, canvas, cell_width, cell_height, num_cells, baseline, was_colored, true, fg, center_glyph);
+    return do_render(self->ct_font, self->units_per_em, bold, italic, info, hb_positions, num_glyphs, canvas, cell_width, cell_height, num_cells, baseline, was_colored, true, fg, center_glyph);
 }
 
 
@@ -691,8 +767,8 @@ postscript_name_for_face(const PyObject *face_) {
 static PyObject *
 repr(CTFace *self) {
     char buf[1024] = {0};
-    snprintf(buf, sizeof(buf)/sizeof(buf[0]), "ascent=%.1f, descent=%.1f, leading=%.1f, point_sz=%.1f, scaled_point_sz=%.1f, underline_position=%.1f underline_thickness=%.1f",
-        (self->ascent), (self->descent), (self->leading), (self->point_sz), (self->scaled_point_sz), (self->underline_position), (self->underline_thickness));
+    snprintf(buf, sizeof(buf)/sizeof(buf[0]), "ascent=%.1f, descent=%.1f, leading=%.1f, scaled_point_sz=%.1f, underline_position=%.1f underline_thickness=%.1f",
+        (self->ascent), (self->descent), (self->leading), (self->scaled_point_sz), (self->underline_position), (self->underline_thickness));
     return PyUnicode_FromFormat(
         "Face(family=%U, full_name=%U, postscript_name=%U, path=%U, units_per_em=%u, %s)",
         self->family_name, self->full_name, self->postscript_name, self->path, self->units_per_em, buf
@@ -708,7 +784,6 @@ static PyMethodDef module_methods[] = {
 static PyMemberDef members[] = {
 #define MEM(name, type) {#name, type, offsetof(CTFace, name), READONLY, #name}
     MEM(units_per_em, T_UINT),
-    MEM(point_sz, T_FLOAT),
     MEM(scaled_point_sz, T_FLOAT),
     MEM(ascent, T_FLOAT),
     MEM(descent, T_FLOAT),
