@@ -1,4 +1,4 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>
 
 #
@@ -7,12 +7,10 @@
 #
 
 import math
-from functools import partial as p, wraps
+from functools import lru_cache, wraps
+from functools import partial as p
 from itertools import repeat
-from typing import (
-    Any, Callable, Dict, Iterable, Iterator, List, MutableSequence, Optional,
-    Sequence, Tuple
-)
+from typing import Any, Callable, Dict, Iterable, Iterator, List, Literal, MutableSequence, Optional, Sequence, Tuple
 
 scale = (0.001, 1., 1.5, 2.)
 _dpi = 96.0
@@ -282,6 +280,16 @@ def cross_line(buf: SSByteArray, width: int, height: int, left: bool = True, lev
     thick_line(buf, width, height, buf.supersample_factor * thickness(level), p1, p2)
 
 
+@supersampled()
+def cross_shade(buf: SSByteArray, width: int, height: int, rotate: bool = False, num_of_lines: int = 7) -> None:
+    line_thickness = max(buf.supersample_factor, width // num_of_lines)
+    delta = int(2 * line_thickness)
+    y1, y2 = (height, 0) if rotate else (0, height)
+    for x in range(0, width, delta):
+        thick_line(buf, width, height, line_thickness, (0 + x, y1), (width + x, y2))
+        thick_line(buf, width, height, line_thickness, (0 - x, y1), (width - x, y2))
+
+
 @supersampled()
 def half_cross_line(buf: SSByteArray, width: int, height: int, which: str = 'tl', level: int = 1) -> None:
     thickness_in_pixels = thickness(level) * buf.supersample_factor
@@ -435,13 +443,31 @@ def draw_parametrized_curve(
                         buf[pos] = min(255, buf[pos] + 255)
 
 
+def circle_equations(
+    origin_x: int = 0, origin_y: int = 0, radius: float = 10., # radius is in pixels as are origin co-ords
+    start_at: float = 0., end_at: float = 360.
+) -> Tuple[ParameterizedFunc, ParameterizedFunc]:
+    conv = math.pi / 180.
+    start = start_at * conv
+    end = end_at * conv
+    amt = end - start
+
+    def x(t: float) -> float:
+        return origin_x + radius * math.cos(start + amt * t)
+
+    def y(t: float) -> float:
+        return origin_y + radius * math.sin(start + amt * t)
+
+    return x, y
+
+
 def rectircle_equations(
     cell_width: int, cell_height: int, supersample_factor: int,
     which: str = '╭'
 ) -> Tuple[ParameterizedFunc, ParameterizedFunc]:
     '''
     Return two functions, x(t) and y(t) that map the parameter t which must be
-    in the range [0, 1] to x and y co-ordinates in the cell. The rectircle equation
+    in the range [0, 1] to x and y coordinates in the cell. The rectircle equation
     we use is:
 
     (|x| / a) ^ (2a / r) + (|y| / a) ^ (2b / r) = 1
@@ -514,6 +540,14 @@ def rounded_separator(buf: SSByteArray, width: int, height: int, level: int = 1,
                 buf[offset + dest_x] = mbuf[offset + src_x]
 
 
+@supersampled()
+def spinner(buf: SSByteArray, width: int, height: int, level: int = 1, start: float = 0, end: float = 360) -> None:
+    w, h = width // 2, height // 2
+    radius = min(w, h) - int(thickness(level) * buf.supersample_factor) // 2
+    arc_x, arc_y = circle_equations(w, h, radius, start_at=start, end_at=end)
+    draw_parametrized_curve(buf, width, height, level, arc_x, arc_y)
+
+
 def half_dhline(buf: BufType, width: int, height: int, level: int = 1, which: str = 'left', only: Optional[str] = None) -> Tuple[int, int]:
     x1, x2 = (0, width // 2) if which == 'left' else (width // 2, width)
     gap = thickness(level + 1, horizontal=False)
@@ -615,41 +649,66 @@ def inner_corner(buf: BufType, width: int, height: int, which: str = 'tl', level
     draw_vline(buf, width, y1, y2, width // 2 + (xd * hgap), level)
 
 
-def shade(buf: BufType, width: int, height: int, light: bool = False, invert: bool = False) -> None:
-    square_sz = max(1, width // 12)
-    number_of_rows = height // square_sz
-    number_of_cols = width // square_sz
-    nums = tuple(range(square_sz))
+def shade(
+    buf: BufType, width: int, height: int, light: bool = False, invert: bool = False, which_half: str = '', fill_blank: bool = False,
+    xnum: int = 12, ynum: int = 0,
+) -> None:
+    square_width = max(1, width // xnum)
+    square_height = max(1, (height // ynum) if ynum else square_width)
+    number_of_rows = height // square_height
+    number_of_cols = width // square_width
+    rows = range(number_of_rows)
+    cols = range(number_of_cols)
+    if which_half == 'top':
+        rows = range(number_of_rows//2)
+    elif which_half == 'bottom':
+        rows = range(number_of_rows//2, number_of_rows)
+    elif which_half == 'left':
+        cols = range(number_of_cols // 2)
+    elif which_half == 'right':
+        cols = range(number_of_cols // 2, number_of_cols)
 
-    dest = bytearray(width * height) if invert else buf
+    for r in rows:
+        for c in cols:
+            if invert ^ ((r % 2 != c % 2) or (light and r % 2 == 1)):
+                continue
+            for yr in range(square_height):
+                y = r * square_height + yr
+                offset = width * y
+                for xc in range(square_width):
+                    x = c * square_width + xc
+                    buf[offset + x] = 255
+    if not fill_blank:
+        return
+    if which_half == 'bottom':
+        rows = range(height//2)
+        cols = range(width)
+    elif which_half == 'top':
+        rows = range(height//2 - 1, height)
+        cols = range(width)
+    elif which_half == 'right':
+        cols = range(width // 2)
+        rows = range(height)
+    elif which_half == 'left':
+        cols = range(width // 2 - 1, width)
+        rows = range(height)
 
-    for r in range(number_of_rows):
-        is_odd = r % 2 != 0
-        if is_odd:
-            continue
-        fill_even = r % 4 == 0
-        for yr in nums:
-            y = r * square_sz + yr
-            if y >= height:
-                break
-            off = width * y
-            for c in range(number_of_cols):
-                if light:
-                    fill = (c % 4) == (0 if fill_even else 2)
-                else:
-                    fill = (c % 2 == 0) == fill_even
-                if fill:
-                    for xc in nums:
-                        x = (c * square_sz) + xc
-                        if x >= width:
-                            break
-                        dest[off + x] = 255
-    if invert:
-        for y in range(height):
-            off = width * y
-            for x in range(width):
-                q = off + x
-                buf[q] = 255 - dest[q]
+    for r in rows:
+        off = r * width
+        for c in cols:
+            buf[off + c] = 255
+
+
+def mask(
+    mask_func: Callable[[BufType, int, int], None], buf: BufType, width: int, height: int,
+) -> None:
+    m = bytearray(width * height)
+    mask_func(m, width, height)
+    for y in range(height):
+        offset = y * width
+        for x in range(width):
+            p = offset + x
+            buf[p] = int(255.0 * (buf[p] / 255.0 * m[p] / 255.0))
 
 
 def quad(buf: BufType, width: int, height: int, x: int = 0, y: int = 0) -> None:
@@ -754,19 +813,83 @@ def eight_block(buf: BufType, width: int, height: int, level: int = 1, which: Tu
         eight_bar(buf, width, height, level, x, horizontal)
 
 
+def frame(buf: BufType, width: int, height: int, edges: Tuple[Literal['l', 'r', 't', 'b'], ...] = ('l', 'r', 't', 'b'), level: int = 0) -> None:
+    h = thickness(level=level, horizontal=True)
+    v = thickness(level=level, horizontal=False)
+
+    def line(x1: int, x2: int, y1: int, y2: int) -> None:
+        for y in range(y1, y2):
+            offset = y * width
+            for x in range(x1, x2):
+                buf[x + offset] = 255
+
+    def hline(y1: int, y2: int) -> None:
+        line(0, width, y1, y2)
+
+    def vline(x1: int, x2: int) -> None:
+        line(x1, x2, 0, height)
+
+    if 't' in edges:
+        hline(0, h + 1)
+    if 'b' in edges:
+        hline(height - h - 1, height)
+    if 'l' in edges:
+        vline(0, v + 1)
+    if 'r' in edges:
+        vline(width - v - 1, width)
+
+
+def progress_bar(buf: BufType, width: int, height: int, which: Literal['l', 'm', 'r'] = 'l', filled: bool = False, level: int = 1, gap_factor: int = 3) -> None:
+    if which == 'l':
+        frame(buf, width, height, edges=('l', 't', 'b'), level=level)
+    elif which == 'm':
+        frame(buf, width, height, edges=('t', 'b'), level=level)
+    else:
+        frame(buf, width, height, edges=('r', 't', 'b'), level=level)
+    if not filled:
+        return
+    h = thickness(level=level, horizontal=True)
+    v = thickness(level=level, horizontal=False)
+    y1 = gap_factor * h
+    y2 = height - gap_factor*h
+    if which == 'l':
+        x1, x2 = gap_factor * v, width
+    elif which == 'm':
+        x1, x2 = 0, width
+    else:
+        x1, x2 = 0, width - gap_factor*v
+    for y in range(y1, y2):
+        offset = y * width
+        for x in range(x1, x2):
+            buf[x + offset] = 255
+
+
+@lru_cache(maxsize=64)
+def distribute_dots(available_space: int, num_of_dots: int) -> Tuple[Tuple[int, ...], int]:
+    dot_size = max(1, available_space // (2 * num_of_dots))
+    extra = available_space - 2 * num_of_dots * dot_size
+    gaps = list(repeat(dot_size, num_of_dots))
+    if extra > 0:
+        idx = 0
+        while extra > 0:
+            gaps[idx] += 1
+            idx = (idx + 1) % len(gaps)
+            extra -= 1
+    gaps[0] //= 2
+    summed_gaps = tuple(sum(gaps[:i + 1]) for i in range(len(gaps)))
+    return summed_gaps, dot_size
+
+
 def braille_dot(buf: BufType, width: int, height: int, col: int, row: int) -> None:
-    dot_height = max(1, height // 8)
-    dot_width = max(1, width // 4)
-    top_margin = (height - 7 * dot_height) // 2
-    left_margin = (width - 3 * dot_width) // 2
-    x_start = left_margin + (col * 2 * dot_width)
-    y_start = top_margin + (row * 2 * dot_height)
-    if y_start < height:
+    x_gaps, dot_width = distribute_dots(width, 2)
+    y_gaps, dot_height = distribute_dots(height, 4)
+    x_start = x_gaps[col] + col * dot_width
+    y_start = y_gaps[row] + row * dot_height
+    if y_start < height and x_start < width:
         for y in range(y_start, min(height, y_start + dot_height)):
-            if x_start < width:
-                offset = y * width
-                for x in range(x_start, min(width, x_start + dot_width)):
-                    buf[offset + x] = 255
+            offset = y * width
+            for x in range(x_start, min(width, x_start + dot_width)):
+                buf[offset + x] = 255
 
 
 def braille(buf: BufType, width: int, height: int, which: int = 0) -> None:
@@ -825,6 +948,18 @@ box_chars: Dict[str, List[Callable[[BufType, int, int], Any]]] = {
     '': [p(cross_line, left=False)],
     '': [p(corner_triangle, corner='top-right')],
     '': [cross_line],
+    '': [p(progress_bar, which='l')],
+    '': [p(progress_bar, which='m')],
+    '': [p(progress_bar, which='r')],
+    '': [p(progress_bar, which='l', filled=True)],
+    '': [p(progress_bar, which='m', filled=True)],
+    '': [p(progress_bar, which='r', filled=True)],
+    '': [p(spinner, start=235, end=305)],
+    '': [p(spinner, start=270, end=390)],
+    '': [p(spinner, start=315, end=470)],
+    '': [p(spinner, start=360, end=540)],
+    '': [p(spinner, start=80, end=220)],
+    '': [p(spinner, start=170, end=270)],
     '═': [dhline],
     '║': [dvline],
 
@@ -870,9 +1005,29 @@ box_chars: Dict[str, List[Callable[[BufType, int, int], Any]]] = {
     '▎': [p(eight_block, which=(0, 1))],
     '▏': [p(eight_bar)],
     '▐': [p(eight_block, which=(4, 5, 6, 7))],
+
     '░': [p(shade, light=True)],
     '▒': [shade],
-    '▓': [p(shade, invert=True)],
+    '▓': [p(shade, light=True, invert=True)],
+    '🮌': [p(shade, which_half='left')],
+    '🮍': [p(shade, which_half='right')],
+    '🮎': [p(shade, which_half='top')],
+    '🮏': [p(shade, which_half='bottom')],
+    '🮐': [p(shade, invert=True)],
+    '🮑': [p(shade, which_half='bottom', invert=True, fill_blank=True)],
+    '🮒': [p(shade, which_half='top', invert=True, fill_blank=True)],
+    '🮓': [p(shade, which_half='right', invert=True, fill_blank=True)],
+    '🮔': [p(shade, which_half='left', invert=True, fill_blank=True)],
+    '🮕': [p(shade, xnum=4, ynum=4)],
+    '🮖': [p(shade, xnum=4, ynum=4, invert=True)],
+    '🮗': [p(shade, xnum=1, ynum=4, invert=True)],
+    '🮜': [shade, p(mask, p(corner_triangle, corner='top-left'))],
+    '🮝': [shade, p(mask, p(corner_triangle, corner='top-right'))],
+    '🮞': [shade, p(mask, p(corner_triangle, corner='bottom-right'))],
+    '🮟': [shade, p(mask, p(corner_triangle, corner='bottom-left'))],
+    '🮘': [cross_shade],
+    '🮙': [p(cross_shade, rotate=True)],
+
     '▔': [p(eight_bar, horizontal=True)],
     '▕': [p(eight_bar, which=7)],
     '▖': [p(quad, y=1)],
@@ -943,9 +1098,11 @@ box_chars: Dict[str, List[Callable[[BufType, int, int], Any]]] = {
     '🭪': [p(half_triangle, which='right', inverted=True)],
     '🭫': [p(half_triangle, which='bottom', inverted=True)],
     '🭬': [half_triangle],
+    '🮛': [half_triangle, p(half_triangle, which='right')],
     '🭭': [p(half_triangle, which='top')],
     '🭮': [p(half_triangle, which='right')],
     '🭯': [p(half_triangle, which='bottom')],
+    '🮚': [p(half_triangle, which='bottom'), p(half_triangle, which='top')],
 
     '🭼': [eight_bar, p(eight_bar, which=7, horizontal=True)],
     '🭽': [eight_bar, p(eight_bar, horizontal=True)],
@@ -1033,12 +1190,7 @@ def render_box_char(ch: str, buf: BufType, width: int, height: int, dpi: float =
 
 
 def render_missing_glyph(buf: BufType, width: int, height: int) -> None:
-    hgap = thickness(level=0, horizontal=True) + 1
-    vgap = thickness(level=0, horizontal=False) + 1
-    draw_hline(buf, width, hgap, width - hgap + 1, vgap, 0)
-    draw_hline(buf, width, hgap, width - hgap + 1, height - vgap, 0)
-    draw_vline(buf, width, vgap, height - vgap + 1, hgap, 0)
-    draw_vline(buf, width, vgap, height - vgap + 1, width - hgap, 0)
+    frame(buf, width, height)
 
 
 def test_char(ch: str, sz: int = 48) -> None:
