@@ -1044,7 +1044,7 @@ native_window_handle(GLFWwindow *w) {
     void *ans = glfwGetCocoaWindow(w);
     return PyLong_FromVoidPtr(ans);
 #endif
-    if (glfwGetX11Window) return PyLong_FromLong((long)glfwGetX11Window(w));
+    if (glfwGetX11Window) return PyLong_FromUnsignedLong(glfwGetX11Window(w));
     return Py_None;
 }
 
@@ -1077,7 +1077,10 @@ create_os_window(PyObject UNUSED *self, PyObject *args, PyObject *kw) {
         // Request SRGB output buffer
         // Prevents kitty from starting on Wayland + NVIDIA, sigh: https://github.com/kovidgoyal/kitty/issues/7021
         // Remove after https://github.com/NVIDIA/egl-wayland/issues/85 is fixed.
-        if (!global_state.is_wayland || !is_nvidia_gpu_driver()) glfwWindowHint(GLFW_SRGB_CAPABLE, true);
+        // Also apparently mesa has introduced a bug with sRGB surfaces and Wayland.
+        // Sigh. Wayland is such a pile of steaming crap.
+        // See https://github.com/kovidgoyal/kitty/issues/7174#issuecomment-2000033873
+        if (!global_state.is_wayland) glfwWindowHint(GLFW_SRGB_CAPABLE, true);
 #ifdef __APPLE__
         cocoa_set_activation_policy(OPT(macos_hide_from_tasks));
         glfwWindowHint(GLFW_COCOA_GRAPHICS_SWITCHING, true);
@@ -1768,7 +1771,7 @@ x11_window_id(PyObject UNUSED *self, PyObject *os_wid) {
     OSWindow *w = os_window_for_id(PyLong_AsUnsignedLongLong(os_wid));
     if (!w) { PyErr_SetString(PyExc_ValueError, "No OSWindow with the specified id found"); return NULL; }
     if (!glfwGetX11Window) { PyErr_SetString(PyExc_RuntimeError, "Failed to load glfwGetX11Window"); return NULL; }
-    return Py_BuildValue("l", (long)glfwGetX11Window(w->handle));
+    return PyLong_FromUnsignedLong(glfwGetX11Window(w->handle));
 }
 
 static PyObject*
