@@ -236,7 +236,7 @@ cell_for_pos(Window *w, unsigned int *x, unsigned int *y, bool *in_left_half_of_
     bool in_left_half = true;
     double mouse_x = global_state.callback_os_window->mouse_x;
     double mouse_y = global_state.callback_os_window->mouse_y;
-    double left = window_left(w), top = window_top(w), right = window_right(w), bottom = window_bottom(w);
+    double left = g->left, top = g->top, right = g->right, bottom = g->bottom;
     w->mouse_pos.global_x = mouse_x - left; w->mouse_pos.global_y = mouse_y - top;
     if (clamp_to_window) {
         mouse_x = MIN(MAX(mouse_x, left), right);
@@ -384,7 +384,7 @@ HANDLER(handle_move_event) {
     bool cell_half_changed = false;
     if (!set_mouse_position(w, &mouse_cell_changed, &cell_half_changed)) return;
     Screen *screen = w->render_data.screen;
-    if(OPT(detect_urls)) detect_url(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y);
+    if (OPT(detect_urls)) detect_url(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y);
     bool in_tracking_mode = (
         screen->modes.mouse_tracking_mode == ANY_MODE ||
         (screen->modes.mouse_tracking_mode == MOTION_MODE && button >= 0));
@@ -394,7 +394,7 @@ HANDLER(handle_move_event) {
     } else {
         if (!mouse_cell_changed && screen->modes.mouse_tracking_protocol != SGR_PIXEL_PROTOCOL) return;
         int sz = encode_mouse_button(w, button, button >=0 ? DRAG : MOVE, modifiers);
-        if (sz > 0) { mouse_event_buf[sz] = 0; write_escape_code_to_child(screen, CSI, mouse_event_buf); }
+        if (sz > 0) { mouse_event_buf[sz] = 0; write_escape_code_to_child(screen, ESC_CSI, mouse_event_buf); }
     }
 }
 
@@ -534,7 +534,7 @@ dispatch_possible_click(Window *w, int button, int modifiers) {
     Screen *screen = w->render_data.screen;
     int count = multi_click_count(w, button);
     if (release_is_click(w, button)) {
-        PendingClick *pc = calloc(sizeof(PendingClick), 1);
+        PendingClick *pc = calloc(1, sizeof(PendingClick));
         if (pc) {
             const ClickQueue *q = &w->click_queues[button];
             pc->press_num = q->length ? q->clicks[q->length - 1].num : 0;
@@ -566,7 +566,7 @@ HANDLER(handle_button_event) {
     if (!dispatch_mouse_event(w, button, is_release ? -1 : 1, modifiers, screen->modes.mouse_tracking_mode != 0)) {
         if (screen->modes.mouse_tracking_mode != 0) {
             int sz = encode_mouse_button(w, button, is_release ? RELEASE : PRESS, modifiers);
-            if (sz > 0) { mouse_event_buf[sz] = 0; write_escape_code_to_child(screen, CSI, mouse_event_buf); }
+            if (sz > 0) { mouse_event_buf[sz] = 0; write_escape_code_to_child(screen, ESC_CSI, mouse_event_buf); }
         }
     }
     // the windows array might have been re-alloced in dispatch_mouse_event
@@ -957,7 +957,7 @@ scroll_event(double xoffset, double yoffset, int flags, int modifiers) {
                 if (sz > 0) {
                     mouse_event_buf[sz] = 0;
                     for (s = abs(s); s > 0; s--) {
-                        write_escape_code_to_child(screen, CSI, mouse_event_buf);
+                        write_escape_code_to_child(screen, ESC_CSI, mouse_event_buf);
                     }
                 }
             } else {
@@ -974,7 +974,7 @@ scroll_event(double xoffset, double yoffset, int flags, int modifiers) {
                 if (sz > 0) {
                     mouse_event_buf[sz] = 0;
                     for (s = abs(s); s > 0; s--) {
-                        write_escape_code_to_child(screen, CSI, mouse_event_buf);
+                        write_escape_code_to_child(screen, ESC_CSI, mouse_event_buf);
                     }
                 }
             }
@@ -999,7 +999,7 @@ send_mouse_event(PyObject *self UNUSED, PyObject *args, PyObject *kw) {
         int sz = encode_mouse_event_impl(&mpos, screen->modes.mouse_tracking_protocol, button, action, mods);
         if (sz > 0) {
             mouse_event_buf[sz] = 0;
-            write_escape_code_to_child(screen, CSI, mouse_event_buf);
+            write_escape_code_to_child(screen, ESC_CSI, mouse_event_buf);
             Py_RETURN_TRUE;
         }
     }
