@@ -1,4 +1,4 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import base64
@@ -11,26 +11,39 @@ from functools import lru_cache, partial
 from time import monotonic, time_ns
 from types import GeneratorType
 from typing import (
-    TYPE_CHECKING, Any, Dict, FrozenSet, Iterable, Iterator, List, Optional,
-    Tuple, Union, cast
+    TYPE_CHECKING,
+    Any,
+    Dict,
+    FrozenSet,
+    Iterable,
+    Iterator,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+    cast,
 )
 
-from .cli import emph, parse_args
+from .cli import parse_args
 from .cli_stub import RCOptions
 from .constants import RC_ENCRYPTION_PROTOCOL_VERSION, appname, version
 from .fast_data_types import (
-    AES256GCMDecrypt, AES256GCMEncrypt, EllipticCurveKey, get_boss,
-    get_options, read_command_response, send_data_to_peer
-)
-from .rc.base import (
-    NoResponse, ParsingOfArgsFailed, PayloadGetter, all_command_names,
-    command_for_name, parse_subcommand_cli
+    AES256GCMDecrypt,
+    AES256GCMEncrypt,
+    EllipticCurveKey,
+    get_boss,
+    get_options,
+    read_command_response,
+    send_data_to_peer,
 )
+from .rc.base import NoResponse, PayloadGetter, all_command_names, command_for_name
 from .types import AsyncResponse
 from .typing import BossType, WindowType
 from .utils import TTYIO, log_error, parse_address_spec, resolve_custom_file
 
 active_async_requests: Dict[str, float] = {}
+active_streams: Dict[str, str] = {}
 if TYPE_CHECKING:
     from .window import Window
 
@@ -39,12 +52,16 @@ def encode_response_for_peer(response: Any) -> bytes:
     return b'\x1bP@kitty-cmd' + json.dumps(response).encode('utf-8') + b'\x1b\\'
 
 
-def parse_cmd(serialized_cmd: str, encryption_key: EllipticCurveKey) -> Dict[str, Any]:
+def parse_cmd(serialized_cmd: memoryview, encryption_key: EllipticCurveKey) -> Dict[str, Any]:
+    # See https://github.com/python/cpython/issues/74379 for why we cant use
+    # memoryview directly :((
     try:
-        pcmd = json.loads(serialized_cmd)
+        pcmd = json.loads(bytes(serialized_cmd))
     except Exception:
+        log_error('Failed to parse JSON payload of remote command, ignoring it')
         return {}
     if not isinstance(pcmd, dict) or 'version' not in pcmd:
+        log_error('JSON payload of remote command is invalid, must be an object with a version field')
         return {}
     pcmd.pop('password', None)
     if 'encrypted' in pcmd:
@@ -92,6 +109,27 @@ def fnmatch_pattern(pat: str) -> 're.Pattern[str]':
     return re.compile(translate(pat))
 
 
+def remote_control_allowed(
+    pcmd: Dict[str, Any], remote_control_passwords: Optional[Dict[str, Sequence[str]]],
+    window: Optional['Window'], extra_data: Dict[str, Any]
+) -> bool:
+    if not remote_control_passwords:
+        return True
+    pw = pcmd.get('password', '')
+    auth_items = remote_control_passwords.get(pw)
+    if pw == '!':
+        auth_items = None
+    if auth_items is None:
+        if '!' in remote_control_passwords:
+            raise PermissionError()
+        return False
+    from .remote_control import password_authorizer
+    pa = password_authorizer(auth_items)
+    if not pa.is_cmd_allowed(pcmd, window, False, extra_data):
+        raise PermissionError()
+    return True
+
+
 class PasswordAuthorizer:
 
     def __init__(self, auth_items: FrozenSet[str]) -> None:
@@ -136,6 +174,16 @@ user_password_allowed: Dict[str, bool] = {}
 
 
 def is_cmd_allowed(pcmd: Dict[str, Any], window: Optional['Window'], from_socket: bool, extra_data: Dict[str, Any]) -> Optional[bool]:
+    sid = pcmd.get('stream_id', '')
+    if sid and active_streams.get(sid, '') == pcmd['cmd']:
+        return True
+    if 'cancel_async' in pcmd and pcmd.get('async_id'):
+        # we allow these without authentication as they are sent on error
+        # conditions and we can't have users prompted for these. The worst side
+        # effect of a malicious cancel_async request is that it can prevent
+        # another async request from getting a result, if it knows the async_id
+        # of that request.
+        return True
     pw = pcmd.get('password', '')
     if not pw:
         auth_items = get_options().remote_control_password.get('')
@@ -157,7 +205,13 @@ def set_user_password_allowed(pwd: str, allowed: bool = True) -> None:
     user_password_allowed[pwd] = allowed
 
 
-def handle_cmd(boss: BossType, window: Optional[WindowType], cmd: Dict[str, Any], peer_id: int) -> Union[Dict[str, Any], None, AsyncResponse]:
+def close_active_stream(stream_id: str) -> None:
+    active_streams.pop(stream_id, None)
+
+
+def handle_cmd(
+    boss: BossType, window: Optional[WindowType], cmd: Dict[str, Any], peer_id: int, self_window: Optional[WindowType]
+) -> Union[Dict[str, Any], None, AsyncResponse]:
     v = cmd['version']
     no_response = cmd.get('no_response', False)
     if tuple(v)[:2] > version[:2]:
@@ -168,18 +222,28 @@ def handle_cmd(boss: BossType, window: Optional[WindowType], cmd: Dict[str, Any]
     payload = cmd.get('payload') or {}
     payload['peer_id'] = peer_id
     async_id = str(cmd.get('async', ''))
+    stream_id = str(cmd.get('stream_id', ''))
+    stream = bool(cmd.get('stream', False))
+    if (stream or stream_id) and not c.reads_streaming_data:
+        return {'ok': False, 'error': 'Streaming send of data is not supported for this command'}
+    if stream_id:
+        payload['stream_id'] = stream_id
+        active_streams[stream_id] = cmd['cmd']
+        if len(active_streams) > 32:
+            oldest = next(iter(active_streams))
+            del active_streams[oldest]
     if async_id:
         payload['async_id'] = async_id
         if 'cancel_async' in cmd:
             active_async_requests.pop(async_id, None)
-            c.cancel_async_request(boss, window, PayloadGetter(c, payload))
+            c.cancel_async_request(boss, self_window or window, PayloadGetter(c, payload))
             return None
         active_async_requests[async_id] = monotonic()
         if len(active_async_requests) > 32:
             oldest = next(iter(active_async_requests))
             del active_async_requests[oldest]
     try:
-        ans = c.response_from_kitty(boss, window, PayloadGetter(c, payload))
+        ans = c.response_from_kitty(boss, self_window or window, PayloadGetter(c, payload))
     except Exception:
         if no_response:  # don't report errors if --no-response was used
             return None
@@ -187,21 +251,25 @@ def handle_cmd(boss: BossType, window: Optional[WindowType], cmd: Dict[str, Any]
     if isinstance(ans, NoResponse):
         return None
     if isinstance(ans, AsyncResponse):
+        if stream:
+            return {'ok': True, 'stream': True}
         return ans
     response: Dict[str, Any] = {'ok': True}
     if ans is not None:
         response['data'] = ans
-    if not c.no_response and not no_response:
+    if not no_response:
         return response
     return None
 
 
 global_options_spec = partial('''\
 --to
-An address for the kitty instance to control. Corresponds to the address
-given to the kitty instance via the :option:`kitty --listen-on` option. If not specified,
-messages are sent to the controlling terminal for this process, i.e. they
-will only work if this process is run within an existing kitty window.
+An address for the kitty instance to control. Corresponds to the address given
+to the kitty instance via the :option:`kitty --listen-on` option or the
+:opt:`listen_on` setting in :file:`kitty.conf`. If not specified, the
+environment variable :envvar:`KITTY_LISTEN_ON` is checked. If that is also not
+found, messages are sent to the controlling terminal for this process, i.e.
+they will only work if this process is run within a kitty window.
 
 
 --password
@@ -215,15 +283,15 @@ completion=type:file relative:conf kwds:-
 default=rc-pass
 A file from which to read the password. Trailing whitespace is ignored. Relative
 paths are resolved from the kitty configuration directory. Use - to read from STDIN.
-Used if no :option:`kitty @ --password` is supplied. Defaults to checking for the
+Used if no :option:`kitten @ --password` is supplied. Defaults to checking for the
 :file:`rc-pass` file in the kitty configuration directory.
 
 
 --password-env
 default=KITTY_RC_PASSWORD
 The name of an environment variable to read the password from.
-Used if no :option:`kitty @ --password-file` is supplied. Defaults
-to checking the :envvar:`KITTY_RC_PASSWORD`.
+Used if no :option:`kitten @ --password-file` is supplied. Defaults
+to checking the environment variable :envvar:`KITTY_RC_PASSWORD`.
 
 
 --use-password
@@ -394,8 +462,7 @@ def create_basic_command(name: str, payload: Any = None, no_response: bool = Fal
 
 
 def send_response_to_client(data: Any = None, error: str = '', peer_id: int = 0, window_id: int = 0, async_id: str = '') -> None:
-    ts = active_async_requests.pop(async_id, None)
-    if ts is None:
+    if active_async_requests.pop(async_id, None) is None:
         return
     if error:
         response: Dict[str, Union[bool, int, str]] = {'ok': False, 'error': error}
@@ -453,81 +520,3 @@ def get_pubkey() -> Tuple[str, bytes]:
         raise SystemExit('KITTY_PUBLIC_KEY has unknown version, if you are running on a remote system, update kitty on this system')
     from base64 import b85decode
     return version, b85decode(pubkey)
-
-
-def main(args: List[str]) -> None:
-    global_opts, items = parse_rc_args(args)
-    password = get_password(global_opts)
-    if password:
-        encryption_version, pubkey = get_pubkey()
-        encrypter = CommandEncrypter(pubkey, encryption_version, password)
-    else:
-        encrypter = NoEncryption()
-
-    if not items:
-        from kitty.shell import main as smain
-        smain(global_opts, encrypter)
-        return
-    cmd = items[0]
-    try:
-        c = command_for_name(cmd)
-    except KeyError:
-        raise SystemExit('{} is not a known command. Known commands are: {}'.format(
-            emph(cmd), ', '.join(x.replace('_', '-') for x in all_command_names())))
-    opts, items = parse_subcommand_cli(c, items)
-    try:
-        payload = c.message_to_kitty(global_opts, opts, items)
-    except ParsingOfArgsFailed as err:
-        exit(str(err))
-    no_response = c.no_response
-    if hasattr(opts, 'no_response'):
-        no_response = opts.no_response
-    response_timeout = c.response_timeout
-    if hasattr(opts, 'response_timeout'):
-        response_timeout = opts.response_timeout
-    response_timeout = encrypter.adjust_response_timeout_for_password(response_timeout)
-    send = create_basic_command(cmd, payload=payload, no_response=no_response, is_asynchronous=c.is_asynchronous)
-    listen_on_from_env = False
-    if not global_opts.to and 'KITTY_LISTEN_ON' in os.environ:
-        global_opts.to = os.environ['KITTY_LISTEN_ON']
-        listen_on_from_env = False
-    if global_opts.to:
-        try:
-            parse_address_spec(global_opts.to)
-        except Exception:
-            msg = f'Invalid listen on address: {global_opts.to}'
-            if listen_on_from_env:
-                msg += '. The KITTY_LISTEN_ON environment variable is set incorrectly'
-            exit(msg)
-    import socket
-    try:
-        response = do_io(global_opts.to, send, no_response, response_timeout, encrypter)
-    except (TimeoutError, socket.timeout):
-        send.pop('payload', None)
-        send['cancel_async'] = True
-        try:
-            do_io(global_opts.to, send, True, 10, encrypter)
-        except KeyboardInterrupt:
-            sys.excepthook = lambda *a: print('Interrupted by user', file=sys.stderr)
-            raise
-        except SocketClosed as e:
-            raise SystemExit(str(e))
-        raise SystemExit(f'Timed out after {response_timeout} seconds waiting for response from kitty')
-    except KeyboardInterrupt:
-        sys.excepthook = lambda *a: print('Interrupted by user', file=sys.stderr)
-        raise
-    except FileNotFoundError:
-        raise SystemExit(f'No listen on socket found at: {global_opts.to}')
-    except SocketClosed as e:
-        raise SystemExit(str(e))
-    if no_response:
-        return
-    if not response.get('ok'):
-        if response.get('tb'):
-            print(response['tb'], file=sys.stderr)
-        raise SystemExit(response['error'])
-    data = response.get('data')
-    if data is not None:
-        if c.string_return_is_error and isinstance(data, str):
-            raise SystemExit(data)
-        print(data)
