@@ -116,7 +116,6 @@ last_focused_os_window_id(void) {
     return ans;
 }
 
-
 static id_type
 current_focused_os_window_id(void) {
     for (size_t i = 0; i < global_state.num_os_windows; i++) {
@@ -746,14 +745,13 @@ PYWRAP1(set_options) {
     Py_RETURN_NONE;
 }
 
-PYWRAP1(set_in_sequence_mode) {
-    global_state.in_sequence_mode = PyObject_IsTrue(args);
-    set_ignore_os_keyboard_processing(global_state.in_sequence_mode);
+PYWRAP1(set_ignore_os_keyboard_processing) {
+    set_ignore_os_keyboard_processing(PyObject_IsTrue(args));
     Py_RETURN_NONE;
 }
 
 static void
-init_screen_render_data(OSWindow *osw, const WindowGeometry *g, ScreenRenderData *d) {
+init_window_render_data(OSWindow *osw, const WindowGeometry *g, WindowRenderData *d) {
     d->dx = gl_size(osw->fonts_data->cell_width, osw->viewport_width);
     d->dy = gl_size(osw->fonts_data->cell_height, osw->viewport_height);
     d->xstart = gl_pos_x(g->left, osw->viewport_width);
@@ -761,14 +759,14 @@ init_screen_render_data(OSWindow *osw, const WindowGeometry *g, ScreenRenderData
 }
 
 PYWRAP1(set_tab_bar_render_data) {
-    ScreenRenderData d = {0};
+    WindowRenderData d = {0};
     WindowGeometry g = {0};
     id_type os_window_id;
     PA("KOIIII", &os_window_id, &d.screen, &g.left, &g.top, &g.right, &g.bottom);
     WITH_OS_WINDOW(os_window_id)
         Py_CLEAR(os_window->tab_bar_render_data.screen);
         d.vao_idx = os_window->tab_bar_render_data.vao_idx;
-        init_screen_render_data(os_window, &g, &d);
+        init_window_render_data(os_window, &g, &d);
         os_window->tab_bar_render_data = d;
         Py_INCREF(os_window->tab_bar_render_data.screen);
     END_WITH_OS_WINDOW
@@ -932,14 +930,14 @@ PYWRAP1(set_window_render_data) {
 #define A(name) &(d.name)
 #define B(name) &(g.name)
     id_type os_window_id, tab_id, window_id;
-    ScreenRenderData d = {0};
+    WindowRenderData d = {0};
     WindowGeometry g = {0};
     PA("KKKOIIII", &os_window_id, &tab_id, &window_id, A(screen), B(left), B(top), B(right), B(bottom));
 
     WITH_WINDOW(os_window_id, tab_id, window_id);
         Py_CLEAR(window->render_data.screen);
         d.vao_idx = window->render_data.vao_idx;
-        init_screen_render_data(osw, &g, &d);
+        init_window_render_data(osw, &g, &d);
         window->render_data = d;
         window->geometry = g;
         Py_INCREF(window->render_data.screen);
@@ -1036,6 +1034,8 @@ PYWRAP1(os_window_font_size) {
                 }
             }
             os_window_update_size_increments(os_window);
+            // On Wayland with CSD title needs to be re-rendered in a different font size
+            if (os_window->window_title && global_state.is_wayland) set_os_window_title(os_window, NULL);
         }
         return Py_BuildValue("d", os_window->font_sz_in_pts);
     END_WITH_OS_WINDOW
@@ -1071,6 +1071,26 @@ PYWRAP1(get_os_window_size) {
     Py_RETURN_NONE;
 }
 
+PYWRAP1(get_os_window_pos) {
+    id_type os_window_id;
+    PA("K", &os_window_id);
+    WITH_OS_WINDOW(os_window_id)
+        int x, y;
+        get_os_window_pos(os_window, &x, &y);
+        return Py_BuildValue("ii", x, y);
+    END_WITH_OS_WINDOW
+    Py_RETURN_NONE;
+}
+
+PYWRAP1(set_os_window_pos) {
+    id_type os_window_id;
+    int x, y;
+    PA("Kii", &os_window_id, &x, &y);
+    WITH_OS_WINDOW(os_window_id)
+        set_os_window_pos(os_window, x, y);
+    END_WITH_OS_WINDOW
+    Py_RETURN_NONE;
+}
 
 PYWRAP1(set_boss) {
     Py_CLEAR(global_state.boss);
@@ -1153,15 +1173,8 @@ pyset_background_image(PyObject *self UNUSED, PyObject *args) {
         bgimage = calloc(1, sizeof(BackgroundImage));
         if (!bgimage) return PyErr_NoMemory();
         bool ok;
-        if (png_data) {
-            FILE *fp = fmemopen(png_data, png_data_size, "r");
-            if (fp == NULL) {
-                PyErr_SetFromErrnoWithFilename(PyExc_OSError, path);
-                free(bgimage);
-                return NULL;
-            }
-            ok = png_from_file_pointer(fp, path, &bgimage->bitmap, &bgimage->width, &bgimage->height, &size);
-            fclose(fp);
+        if (png_data && png_data_size) {
+            ok = png_from_data(png_data, png_data_size, path, &bgimage->bitmap, &bgimage->width, &bgimage->height, &size);
         } else {
             ok = png_path_to_bitmap(path, &bgimage->bitmap, &bgimage->width, &bgimage->height, &size);
         }
@@ -1257,6 +1270,18 @@ move_cursor_to_mouse_if_in_prompt(id_type os_window_id, id_type tab_id, id_type
     return moved;
 }
 
+static PyObject*
+pyupdate_pointer_shape(PyObject *self UNUSED, PyObject *args) {
+    id_type os_window_id;
+    PA("K", &os_window_id);
+    WITH_OS_WINDOW(os_window_id);
+    OSWindow *orig = global_state.callback_os_window;
+    global_state.callback_os_window = os_window;
+    update_mouse_pointer_shape();
+    global_state.callback_os_window = orig;
+    END_WITH_OS_WINDOW;
+    Py_RETURN_NONE;
+}
 
 static PyObject*
 pymouse_selection(PyObject *self UNUSED, PyObject *args) {
@@ -1321,10 +1346,26 @@ KKK(set_active_window)
 KII(swap_tabs)
 KK5I(add_borders_rect)
 
+static PyObject*
+os_window_focus_counters(PyObject *self UNUSED, PyObject *args UNUSED) {
+    RAII_PyObject(ans, PyDict_New());
+    for (size_t i = 0; i < global_state.num_os_windows; i++) {
+        OSWindow *w = &global_state.os_windows[i];
+        RAII_PyObject(key, PyLong_FromUnsignedLongLong(w->id));
+        RAII_PyObject(val, PyLong_FromUnsignedLongLong(w->last_focused_counter));
+        if (!key || !val) return NULL;
+        if (PyDict_SetItem(ans, key, val) != 0) return NULL;
+    }
+    Py_INCREF(ans);
+    return ans;
+}
+
 #define M(name, arg_type) {#name, (PyCFunction)name, arg_type, NULL}
 #define MW(name, arg_type) {#name, (PyCFunction)py##name, arg_type, NULL}
 
 static PyMethodDef module_methods[] = {
+    M(os_window_focus_counters, METH_NOARGS),
+    MW(update_pointer_shape, METH_VARARGS),
     MW(current_os_window, METH_NOARGS),
     MW(next_window_id, METH_NOARGS),
     MW(last_focused_os_window_id, METH_NOARGS),
@@ -1337,7 +1378,7 @@ static PyMethodDef module_methods[] = {
     MW(redirect_mouse_handling, METH_O),
     MW(mouse_selection, METH_VARARGS),
     MW(set_window_logo, METH_VARARGS),
-    MW(set_in_sequence_mode, METH_O),
+    MW(set_ignore_os_keyboard_processing, METH_O),
     MW(handle_for_window_id, METH_VARARGS),
     MW(update_ime_position_for_window, METH_VARARGS),
     MW(pt_to_px, METH_VARARGS),
@@ -1372,6 +1413,8 @@ static PyMethodDef module_methods[] = {
     MW(sync_os_window_title, METH_VARARGS),
     MW(get_os_window_title, METH_VARARGS),
     MW(set_os_window_title, METH_VARARGS),
+    MW(get_os_window_pos, METH_VARARGS),
+    MW(set_os_window_pos, METH_VARARGS),
     MW(global_font_size, METH_VARARGS),
     MW(set_background_image, METH_VARARGS),
     MW(os_window_font_size, METH_VARARGS),
@@ -1406,9 +1449,8 @@ finalize(void) {
     free_bgimage(&global_state.bgimage, false);
     free_window_logo_table(&global_state.all_window_logos);
     global_state.bgimage = NULL;
-    free_url_prefixes();
-    free(OPT(select_by_word_characters)); OPT(select_by_word_characters) = NULL;
-    free(OPT(url_excluded_characters)); OPT(url_excluded_characters) = NULL;
+
+    free_allocs_in_options(&global_state.opts);
 }
 
 bool
