@@ -1,4 +1,4 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import json
@@ -11,7 +11,7 @@ from enum import Enum, IntEnum, auto
 from functools import lru_cache, partial
 from gettext import gettext as _
 from itertools import chain
-from time import monotonic
+from time import monotonic, time_ns
 from typing import (
     TYPE_CHECKING,
     Any,
@@ -44,10 +44,13 @@ from .fast_data_types import (
     CURSOR_BEAM,
     CURSOR_BLOCK,
     CURSOR_UNDERLINE,
-    DCS,
+    ESC_DCS,
+    ESC_OSC,
     GLFW_MOD_CONTROL,
+    GLFW_PRESS,
+    GLFW_RELEASE,
+    GLFW_REPEAT,
     NO_CURSOR_SHAPE,
-    OSC,
     SCROLL_FULL,
     SCROLL_LINE,
     SCROLL_PAGE,
@@ -56,6 +59,7 @@ from .fast_data_types import (
     Screen,
     add_timer,
     add_window,
+    base64_decode,
     cell_size_for_window,
     click_mouse_cmd_output,
     click_mouse_url,
@@ -64,15 +68,19 @@ from .fast_data_types import (
     get_boss,
     get_click_interval,
     get_options,
+    is_css_pointer_name_valid,
     last_focused_os_window_id,
     mark_os_window_dirty,
     mouse_selection,
     move_cursor_to_mouse_if_in_prompt,
+    pointer_name_to_css_name,
     pt_to_px,
+    replace_c0_codes_except_nl_space_tab,
     set_window_logo,
     set_window_padding,
     set_window_render_data,
     update_ime_position_for_window,
+    update_pointer_shape,
     update_window_title,
     update_window_visibility,
     wakeup_main_loop,
@@ -80,7 +88,9 @@ from .fast_data_types import (
 from .keys import keyboard_mode_name, mod_mask
 from .notify import (
     NotificationCommand,
+    OnlyWhen,
     handle_notification_cmd,
+    notify_with_command,
     sanitize_identifier_pat,
 )
 from .options.types import Options
@@ -162,6 +172,18 @@ class CwdRequest:
                 set_cwd_in_cmdline(reported_cwd, argv)
                 set_server_args_in_cmdline(server_args, argv, allocate_tty=not run_shell)
                 if env is not None:
+                    # Assume env is coming from a local process so drop env
+                    # vars that can cause issues when set on the remote host
+                    if env.get('KITTY_KITTEN_RUN_MODULE') == 'ssh_askpass':
+                        for k in ('KITTY_KITTEN_RUN_MODULE', 'SSH_ASKPASS', 'SSH_ASKPASS_REQUIRE'):
+                            env.pop(k, None)
+                    for k in (
+                        'HOME', 'USER', 'TEMP', 'TMP', 'TMPDIR', 'PATH', 'PWD', 'OLDPWD', 'KITTY_INSTALLATION_DIR',
+                        'HOSTNAME', 'SSH_AUTH_SOCK', 'SSH_AGENT_PID', 'KITTY_STDIO_FORWARDED',
+                        'KITTY_PUBLIC_KEY', 'TERMINFO', 'XDG_RUNTIME_DIR', 'XDG_VTNR',
+                        'XDG_DATA_DIRS', 'XAUTHORITY', 'EDITOR', 'VISUAL',
+                    ):
+                        env.pop(k, None)
                     set_env_in_cmdline(env, argv, clone=False)
                 return ''
             if not window.child_is_remote and (self.request_type is CwdRequestType.last_reported or window.at_prompt):
@@ -169,14 +191,15 @@ class CwdRequest:
         return window.get_cwd_of_child(oldest=self.request_type is CwdRequestType.oldest) or ''
 
 
-def process_title_from_child(title: str, is_base64: bool) -> str:
+def process_title_from_child(title: memoryview, is_base64: bool, default_title: str) -> str:
     if is_base64:
-        from base64 import standard_b64decode
         try:
-            title = standard_b64decode(title).decode('utf-8', 'replace')
+            stitle = base64_decode(title).decode('utf-8', 'replace')
         except Exception:
-            title = 'undecodeable title'
-    return sanitize_title(title)
+            stitle = 'undecodeable title'
+    else:
+        stitle = str(title, 'utf-8', 'replace')
+    return sanitize_title(stitle or default_title)
 
 
 @lru_cache(maxsize=64)
@@ -206,6 +229,8 @@ class WindowDict(TypedDict):
     lines: int
     columns: int
     user_vars: Dict[str, str]
+    at_prompt: bool
+    created_at: int
 
 
 class PipeData(TypedDict):
@@ -253,11 +278,17 @@ class Watchers:
     on_resize: List[Watcher]
     on_close: List[Watcher]
     on_focus_change: List[Watcher]
+    on_set_user_var: List[Watcher]
+    on_title_change: List[Watcher]
+    on_cmd_startstop: List[Watcher]
 
     def __init__(self) -> None:
         self.on_resize = []
         self.on_close = []
         self.on_focus_change = []
+        self.on_set_user_var = []
+        self.on_title_change = []
+        self.on_cmd_startstop = []
 
     def add(self, others: 'Watchers') -> None:
         def merge(base: List[Watcher], other: List[Watcher]) -> None:
@@ -267,20 +298,28 @@ class Watchers:
         merge(self.on_resize, others.on_resize)
         merge(self.on_close, others.on_close)
         merge(self.on_focus_change, others.on_focus_change)
+        merge(self.on_set_user_var, others.on_set_user_var)
+        merge(self.on_title_change, others.on_title_change)
+        merge(self.on_cmd_startstop, others.on_cmd_startstop)
 
     def clear(self) -> None:
         del self.on_close[:], self.on_resize[:], self.on_focus_change[:]
+        del self.on_set_user_var[:], self.on_title_change[:], self.on_cmd_startstop[:]
 
     def copy(self) -> 'Watchers':
         ans = Watchers()
         ans.on_close = self.on_close[:]
         ans.on_resize = self.on_resize[:]
         ans.on_focus_change = self.on_focus_change[:]
+        ans.on_set_user_var = self.on_set_user_var[:]
+        ans.on_title_change = self.on_title_change[:]
+        ans.on_cmd_startstop = self.on_cmd_startstop[:]
         return ans
 
     @property
     def has_watchers(self) -> bool:
-        return bool(self.on_close or self.on_resize or self.on_focus_change)
+        return bool(self.on_close or self.on_resize or self.on_focus_change
+                    or self.on_set_user_var or self.on_title_change or self.on_cmd_startstop)
 
 
 def call_watchers(windowref: Callable[[], Optional['Window']], which: str, data: Dict[str, Any]) -> None:
@@ -410,12 +449,8 @@ def cmd_output(screen: Screen, which: CommandOutput = CommandOutput.last_run, as
     return ''.join(lines)
 
 
-def process_remote_print(msg: str) -> str:
-    from base64 import standard_b64decode
-
-    from .cli import green
-    text = standard_b64decode(msg).decode('utf-8', 'replace')
-    return text.replace('\x1b', green(r'\e')).replace('\a', green(r'\a')).replace('\0', green(r'\0'))
+def process_remote_print(msg: memoryview) -> str:
+    return replace_c0_codes_except_nl_space_tab(base64_decode(msg)).decode('utf-8', 'replace')
 
 
 class EdgeWidths:
@@ -507,10 +542,12 @@ class Window:
         self.is_focused: bool = False
         self.last_resized_at = 0.
         self.started_at = monotonic()
+        self.created_at = time_ns()
         self.current_remote_data: List[str] = []
         self.current_mouse_event_button = 0
         self.current_clipboard_read_ask: Optional[bool] = None
         self.prev_osc99_cmd = NotificationCommand()
+        self.last_cmd_output_start_time = 0.
         self.actions_on_close: List[Callable[['Window'], None]] = []
         self.actions_on_focus_change: List[Callable[['Window', bool], None]] = []
         self.actions_on_removal: List[Callable[['Window'], None]] = []
@@ -553,21 +590,8 @@ class Window:
     def remote_control_allowed(self, pcmd: Dict[str, Any], extra_data: Dict[str, Any]) -> bool:
         if not self.allow_remote_control:
             return False
-        if not self.remote_control_passwords:
-            return True
-        pw = pcmd.get('password', '')
-        auth_items = self.remote_control_passwords.get(pw)
-        if pw == '!':
-            auth_items = None
-        if auth_items is None:
-            if '!' in self.remote_control_passwords:
-                raise PermissionError()
-            return False
-        from .remote_control import password_authorizer
-        pa = password_authorizer(auth_items)
-        if not pa.is_cmd_allowed(pcmd, self, False, extra_data):
-            raise PermissionError()
-        return True
+        from .remote_control import remote_control_allowed
+        return remote_control_allowed(pcmd, self.remote_control_passwords, self, extra_data)
 
     @property
     def file_transmission_control(self) -> 'FileTransmission':
@@ -585,11 +609,13 @@ class Window:
         self.os_window_id = tab.os_window_id
         self.tabref = weakref.ref(tab)
 
-    def effective_margin(self, edge: EdgeLiteral, is_single_window: bool = False) -> int:
+    def effective_margin(self, edge: EdgeLiteral) -> int:
         q = getattr(self.margin, edge)
         if q is not None:
             return pt_to_px(q, self.os_window_id)
         opts = get_options()
+        tab = self.tabref()
+        is_single_window = tab is not None and tab.has_single_window_visible()
         if is_single_window:
             q = getattr(opts.single_window_margin_width, edge)
             if q > -0.1:
@@ -601,7 +627,14 @@ class Window:
         q = getattr(self.padding, edge)
         if q is not None:
             return pt_to_px(q, self.os_window_id)
-        q = getattr(get_options().window_padding_width, edge)
+        opts = get_options()
+        tab = self.tabref()
+        is_single_window = tab is not None and tab.has_single_window_visible()
+        if is_single_window:
+            q = getattr(opts.single_window_padding_width, edge)
+            if q > -0.1:
+                return pt_to_px(q, self.os_window_id)
+        q = getattr(opts.window_padding_width, edge)
         return pt_to_px(q, self.os_window_id)
 
     def update_effective_padding(self) -> None:
@@ -645,12 +678,14 @@ class Window:
             'pid': self.child.pid,
             'cwd': self.child.current_cwd or self.child.cwd,
             'cmdline': self.child.cmdline,
-            'env': self.child.environ,
+            'env': self.child.environ or self.child.final_env,
             'foreground_processes': self.child.foreground_processes,
             'is_self': is_self,
+            'at_prompt': self.at_prompt,
             'lines': self.screen.lines,
             'columns': self.screen.columns,
             'user_vars': self.user_vars,
+            'created_at': self.created_at,
         }
 
     def serialize_state(self) -> Dict[str, Any]:
@@ -667,6 +702,7 @@ class Window:
             'env': self.child.environ,
             'cmdline': self.child.cmdline,
             'margin': self.margin.serialize(),
+            'user_vars': self.user_vars,
             'padding': self.padding.serialize(),
         }
         if self.window_custom_type:
@@ -697,9 +733,6 @@ class Window:
         return not self.at_prompt
 
     def matches(self, field: str, pat: MatchPatternType) -> bool:
-        if not pat:
-            return False
-
         if isinstance(pat, tuple):
             if field == 'env':
                 return key_val_matcher(self.child.environ.items(), *pat)
@@ -792,7 +825,7 @@ class Window:
         if self.destroyed:
             return
         if self.needs_layout or new_geometry.xnum != self.screen.columns or new_geometry.ynum != self.screen.lines:
-            self.screen.resize(new_geometry.ynum, new_geometry.xnum)
+            self.screen.resize(max(0, new_geometry.ynum), max(0, new_geometry.xnum))
             self.needs_layout = False
             call_watchers(weakref.ref(self), 'on_resize', {'old_geometry': self.geometry, 'new_geometry': new_geometry})
         current_pty_size = (
@@ -839,6 +872,47 @@ class Window:
         self.write_to_child(text)
         return False
 
+    @ac(
+        'misc', '''
+        Send the specified keys to the active window.
+        Note that the key will be sent only if the current keyboard mode of the program running in the terminal supports it.
+        Both key press and key release are sent. First presses for all specified keys and then releases in reverse order.
+        To send a pattern of press and release for multiple keys use the :ac:`combine` action. For example::
+
+            map f1 send_key ctrl+x alt+y
+            map f1 combine : send_key ctrl+x : send_key alt+y
+    ''')
+    def send_key(self, *args: str) -> bool:
+        from .options.utils import parse_shortcut
+        km = get_options().kitty_mod
+        passthrough = True
+        events = []
+        prev = ''
+        for human_key in args:
+            sk = parse_shortcut(human_key)
+            if sk.is_native:
+                raise ValueError(f'Native key codes not allowed in send_key: {human_key}')
+            sk = sk.resolve_kitty_mod(km)
+            events.append(KeyEvent(key=sk.key, mods=sk.mods, action=GLFW_REPEAT if human_key == prev else GLFW_PRESS))
+            prev = human_key
+        for ev in events + [KeyEvent(key=x.key, mods=x.mods, action=GLFW_RELEASE) for x in reversed(events)]:
+            enc = self.encoded_key(ev)
+            if enc:
+                self.write_to_child(enc)
+                passthrough = False
+        return passthrough
+
+    def send_key_sequence(self, *keys: KeyEvent, synthesize_release_events: bool = True) -> None:
+        for key in keys:
+            enc = self.encoded_key(key)
+            if enc:
+                self.write_to_child(enc)
+            if synthesize_release_events and key.action != GLFW_RELEASE:
+                rkey = KeyEvent(key=key.key, mods=key.mods, action=GLFW_RELEASE)
+                enc = self.encoded_key(rkey)
+                if enc:
+                    self.write_to_child(enc)
+
     @ac('debug', 'Show a dump of the current lines in the scrollback + screen with their line attributes')
     def dump_lines_with_attrs(self) -> None:
         strings: List[str] = []
@@ -863,8 +937,39 @@ class Window:
         if title:
             title = sanitize_title(title)
         self.override_title = title or None
+        self.call_watchers(self.watchers.on_title_change, {'title': self.title, 'from_child': False})
         self.title_updated()
 
+    @ac(
+        'win', '''
+        Change the title of the active window interactively, by typing in the new title.
+        If you specify an argument to this action then that is used as the title instead of asking for it.
+        Use the empty string ("") to reset the title to default. Use a space (" ") to indicate that the
+        prompt should not be pre-filled. For example::
+
+            # interactive usage
+            map f1 set_window_title
+            # set a specific title
+            map f2 set_window_title some title
+            # reset to default
+            map f3 set_window_title ""
+            # interactive usage without prefilled prompt
+            map f3 set_window_title " "
+        '''
+    )
+    def set_window_title(self, title: Optional[str] = None) -> None:
+        if title is not None and title not in ('" "', "' '"):
+            if title in ('""', "''"):
+                title = ''
+            self.set_title(title)
+            return
+        prefilled = self.title
+        if title in ('" "', "' '"):
+            prefilled = ''
+        get_boss().get_line(
+            _('Enter the new title for this window below. An empty title will cause the default title to be used.'),
+            self.set_title, window=self, initial_value=prefilled)
+
     def set_user_var(self, key: str, val: Optional[Union[str, bytes]]) -> None:
         key = sanitize_control_codes(key).replace('\n', ' ')
         self.user_vars.pop(key, None)  # ensure key will be newest in user_vars even if already present
@@ -874,7 +979,10 @@ class Window:
         if val is not None:
             if isinstance(val, bytes):
                 val = val.decode('utf-8', 'replace')
-            self.user_vars[key] = sanitize_control_codes(val).replace('\n', ' ')
+            self.user_vars[key] = val = sanitize_control_codes(val).replace('\n', ' ')
+            self.call_watchers(self.watchers.on_set_user_var, {'key': key, 'value': val})
+        else:
+            self.call_watchers(self.watchers.on_set_user_var, {'key': key, 'value': None})
 
     # screen callbacks {{{
 
@@ -882,11 +990,11 @@ class Window:
         for record in raw_data.split(';'):
             key, _, val = record.partition('=')
             if key == 'SetUserVar':
-                from base64 import standard_b64decode
                 ukey, has_equal, uval = val.partition('=')
-                self.set_user_var(ukey, (standard_b64decode(uval) if uval else b'') if has_equal == '=' else None)
+                self.set_user_var(ukey, (base64_decode(uval) if uval else b'') if has_equal == '=' else None)
 
-    def desktop_notify(self, osc_code: int, raw_data: str) -> None:
+    def desktop_notify(self, osc_code: int, raw_datab: memoryview) -> None:
+        raw_data = str(raw_datab, 'utf-8', 'replace')
         if osc_code == 1337:
             self.osc_1337(raw_data)
         if osc_code == 777:
@@ -898,9 +1006,6 @@ class Window:
         if cmd is not None and osc_code == 99:
             self.prev_osc99_cmd = cmd
 
-    def use_utf8(self, on: bool) -> None:
-        get_boss().child_monitor.set_iutf8_winid(self.id, on)
-
     def on_mouse_event(self, event: Dict[str, Any]) -> bool:
         event['mods'] = event.get('mods', 0) & mod_mask
         ev = MouseEvent(**event)
@@ -935,7 +1040,7 @@ class Window:
                     'What would you like to do with this URL:\n' + styled(sanitize_url_for_dispay_to_user(url), fg='yellow'),
                     partial(self.hyperlink_open_confirmed, url, cwd),
                     'o:Open', 'c:Copy to clipboard', 'n;red:Nothing', default='o',
-                    window=self,
+                    window=self, title=_('Hyperlink activated'),
                 )
                 return
         get_boss().open_url(url, cwd=cwd)
@@ -1001,12 +1106,13 @@ class Window:
             # Cancel IME composition after loses focus
             update_ime_position_for_window(self.id, False, -1)
 
-    def title_changed(self, new_title: Optional[str], is_base64: bool = False) -> None:
-        self.child_title = process_title_from_child(new_title or self.default_title, is_base64)
+    def title_changed(self, new_title: Optional[memoryview], is_base64: bool = False) -> None:
+        self.child_title = process_title_from_child(new_title or memoryview(b''), is_base64, self.default_title)
+        self.call_watchers(self.watchers.on_title_change, {'title': self.child_title, 'from_child': True})
         if self.override_title is None:
             self.title_updated()
 
-    def icon_changed(self, new_icon: object) -> None:
+    def icon_changed(self, new_icon: memoryview) -> None:
         pass  # TODO: Implement this
 
     @property
@@ -1071,15 +1177,20 @@ class Window:
         r |= r << 8
         g |= g << 8
         b |= b << 8
-        self.screen.send_escape_code_to_child(OSC, f'{code};rgb:{r:04x}/{g:04x}/{b:04x}')
+        self.screen.send_escape_code_to_child(ESC_OSC, f'{code};rgb:{r:04x}/{g:04x}/{b:04x}')
 
     def report_notification_activated(self, identifier: str) -> None:
         identifier = sanitize_identifier_pat().sub('', identifier)
-        self.screen.send_escape_code_to_child(OSC, f'99;i={identifier};')
+        self.screen.send_escape_code_to_child(ESC_OSC, f'99;i={identifier};')
 
-    def set_dynamic_color(self, code: int, value: Union[str, bytes]) -> None:
-        if isinstance(value, bytes):
-            value = value.decode('utf-8')
+
+    def set_dynamic_color(self, code: int, value: Union[str, bytes, memoryview] = '') -> None:
+        if isinstance(value, (bytes, memoryview)):
+            value = str(value, 'utf-8', 'replace')
+        if code == 22:
+            ret = set_pointer_shape(self.screen, value, self.os_window_id)
+            if ret:
+                self.screen.send_escape_code_to_child(ESC_OSC, '22:' + ret)
         color_changes: Dict[DynamicColor, Optional[str]] = {}
         for val in value.split(';'):
             w = DYNAMIC_COLOR_CODES.get(code)
@@ -1094,7 +1205,8 @@ class Window:
         if color_changes:
             self.change_colors(color_changes)
 
-    def set_color_table_color(self, code: int, value: str) -> None:
+    def set_color_table_color(self, code: int, bvalue: Optional[memoryview] = None) -> None:
+        value = str(bvalue or b'', 'utf-8', 'replace')
         cp = self.screen.color_profile
         if code == 4:
             changed = False
@@ -1123,22 +1235,24 @@ class Window:
 
     def request_capabilities(self, q: str) -> None:
         for result in get_capabilities(q, get_options()):
-            self.screen.send_escape_code_to_child(DCS, result)
+            self.screen.send_escape_code_to_child(ESC_DCS, result)
 
-    def handle_remote_cmd(self, cmd: str) -> None:
+    def handle_remote_cmd(self, cmd: memoryview) -> None:
         get_boss().handle_remote_cmd(cmd, self)
 
-    def handle_remote_echo(self, msg: str) -> None:
-        from base64 import standard_b64decode
-        data = standard_b64decode(msg)
+    def handle_remote_echo(self, msg: memoryview) -> None:
+        data = base64_decode(msg)
+        # ensure we are not writing any control char back as this can lead to command injection on shell prompts
+        # Any bytes outside the printable ASCII range are removed.
+        data = re.sub(rb'[^ -~]', b'', data)
         self.write_to_child(data)
 
-    def handle_remote_ssh(self, msg: str) -> None:
+    def handle_remote_ssh(self, msg: memoryview) -> None:
         from kittens.ssh.utils import get_ssh_data
         for line in get_ssh_data(msg, f'{os.getpid()}-{self.id}'):
             self.write_to_child(line)
 
-    def handle_kitten_result(self, msg: str) -> None:
+    def handle_kitten_result(self, msg: memoryview) -> None:
         import base64
         self.kitten_result = json.loads(base64.b85decode(msg))
         for processor in self.kitten_result_processors:
@@ -1151,17 +1265,18 @@ class Window:
     def add_kitten_result_processor(self, callback: Callable[['Window', Any], None]) -> None:
         self.kitten_result_processors.append(callback)
 
-    def handle_overlay_ready(self, msg: str) -> None:
+    def handle_overlay_ready(self, msg: memoryview) -> None:
         boss = get_boss()
         tab = boss.tab_for_window(self)
         if tab is not None:
             tab.move_window_to_top_of_group(self)
 
-    def append_remote_data(self, msg: str) -> str:
-        if not msg:
+    def append_remote_data(self, msgb: memoryview) -> str:
+        if not msgb:
             cdata = ''.join(self.current_remote_data)
             self.current_remote_data = []
             return cdata
+        msg = str(msgb, 'utf-8', 'replace')
         num, rest = msg.split(':', 1)
         max_size = get_options().clipboard_max_size * 1024 * 1024
         if num == '0' or sum(map(len, self.current_remote_data)) > max_size:
@@ -1169,13 +1284,13 @@ class Window:
         self.current_remote_data.append(rest)
         return ''
 
-    def handle_remote_edit(self, msg: str) -> None:
+    def handle_remote_edit(self, msg: memoryview) -> None:
         cdata = self.append_remote_data(msg)
         if cdata:
             from .launch import remote_edit
             remote_edit(cdata, self)
 
-    def handle_remote_clone(self, msg: str) -> None:
+    def handle_remote_clone(self, msg: memoryview) -> None:
         cdata = self.append_remote_data(msg)
         if cdata:
             ac = get_options().allow_cloning
@@ -1184,6 +1299,7 @@ class Window:
                     'A program running in this window wants to clone it into another window.'
                     ' Allow it do so, once?'),
                     partial(self.handle_remote_clone_confirmation, cdata), window=self,
+                    title=_('Allow cloning of window?'),
                 )
             elif ac in ('yes', 'y', 'true'):
                 self.handle_remote_clone_confirmation(cdata, True)
@@ -1193,8 +1309,9 @@ class Window:
             from .launch import clone_and_launch
             clone_and_launch(cdata, self)
 
-    def handle_remote_askpass(self, msg: str) -> None:
+    def handle_remote_askpass(self, msgb: memoryview) -> None:
         from .shm import SharedMemory
+        msg = str(msgb, 'utf-8')
         with SharedMemory(name=msg, readonly=True) as shm:
             shm.seek(1)
             data = json.loads(shm.read_data_with_size())
@@ -1222,17 +1339,17 @@ class Window:
         else:
             log_error(f'Ignoring ask request with unknown type: {data["type"]}')
 
-    def handle_remote_print(self, msg: str) -> None:
+    def handle_remote_print(self, msg: memoryview) -> None:
         text = process_remote_print(msg)
         print(text, end='', flush=True)
 
     def send_cmd_response(self, response: Any) -> None:
-        self.screen.send_escape_code_to_child(DCS, '@kitty-cmd' + json.dumps(response))
+        self.screen.send_escape_code_to_child(ESC_DCS, '@kitty-cmd' + json.dumps(response))
 
-    def file_transmission(self, data: str) -> None:
+    def file_transmission(self, data: memoryview) -> None:
         self.file_transmission_control.handle_serialized_command(data)
 
-    def clipboard_control(self, data: str, is_partial: Optional[bool] = False) -> None:
+    def clipboard_control(self, data: memoryview, is_partial: Optional[bool] = False) -> None:
         if is_partial is None:
             self.clipboard_request_manager.parse_osc_5522(data)
         else:
@@ -1243,10 +1360,47 @@ class Window:
             if pop:
                 if self.title_stack:
                     self.child_title = self.title_stack.pop()
+                    self.call_watchers(self.watchers.on_title_change, {'title': self.child_title, 'from_child': True})
                     self.title_updated()
             else:
                 if self.child_title:
                     self.title_stack.append(self.child_title)
+
+    def cmd_output_marking(self, is_start: bool) -> None:
+        if is_start:
+            start_time = monotonic()
+            self.last_cmd_output_start_time = start_time
+
+            self.call_watchers(self.watchers.on_cmd_startstop, {"is_start": True, "time": start_time})
+        else:
+            if self.last_cmd_output_start_time > 0.:
+                end_time = monotonic()
+                last_cmd_output_duration = end_time - self.last_cmd_output_start_time
+                self.last_cmd_output_start_time = 0.
+
+                self.call_watchers(self.watchers.on_cmd_startstop, {"is_start": False, "time": end_time})
+
+                opts = get_options()
+                when, duration, action, cmdline = opts.notify_on_cmd_finish
+
+                if last_cmd_output_duration >= duration and when != 'never':
+                    cmd = NotificationCommand()
+                    cmd.title = 'kitty'
+                    cmd.body = 'Command finished in a background window.\nClick to focus.'
+                    cmd.actions = 'focus'
+                    cmd.only_when = OnlyWhen(when)
+                    if action == 'notify':
+                        notify_with_command(cmd, self.id)
+                    elif action == 'bell':
+                        def bell(title: str, body: str, identifier: str) -> None:
+                            self.screen.bell()
+                        notify_with_command(cmd, self.id, notify_implementation=bell)
+                    elif action == 'command':
+                        def run_command(title: str, body: str, identifier: str) -> None:
+                            open_cmd(cmdline)
+                        notify_with_command(cmd, self.id, notify_implementation=run_command)
+                    else:
+                        raise ValueError(f'Unknown action in option `notify_on_cmd_finish`: {action}')
     # }}}
 
     # mouse actions {{{
@@ -1444,20 +1598,45 @@ class Window:
                 if rest.startswith('//') or scheme in ('mailto', 'irc'):
                     import shlex
                     text = shlex.quote(text)
+        if 'replace-dangerous-control-codes' in opts.paste_actions:
+            text = replace_c0_codes_except_nl_space_tab(text)
+        if 'replace-newline' in opts.paste_actions:
+            text = text.replace('\n', '\x1bE')
         btext = text.encode('utf-8')
         if 'confirm' in opts.paste_actions:
-            msg = ''
-            limit = 16 * 1024
+            sanitized = replace_c0_codes_except_nl_space_tab(btext)
             if not self.screen.in_bracketed_paste_mode:
-                msg = _('Pasting text into shells that do not support bracketed paste can be dangerous.')
-            elif len(btext) > limit:
+                # \n is converted to \r and \r is interpreted as the enter key
+                # by legacy programs that dont support the full kitty keyboard protocol,
+                # which in the case of shells can lead to command execution, so
+                # replace with <ESC>E (NEL) which has the newline visual effect \r\n but
+                # isnt interpreted as Enter.
+                sanitized = sanitized.replace(b'\n', b'\x1bE')
+            if sanitized != btext:
+                msg = _('The text to be pasted contains terminal control codes.\n\nIf the terminal program you are pasting into does not properly'
+                        ' sanitize pasted text, this can lead to \x1b[31mcode execution vulnerabilities\x1b[39m.\n\nHow would you like to proceed?')
+                get_boss().choose(
+                    msg, partial(self.handle_dangerous_paste_confirmation, btext, sanitized),
+                    's;green:Sanitize and paste', 'p;red:Paste anyway', 'c;yellow:Cancel',
+                    window=self, default='s', title=_('Allow paste?'),
+                )
+                return
+        if 'confirm-if-large' in opts.paste_actions:
+            msg = ''
+            if len(btext) > 16 * 1024:
                 msg = _('Pasting very large amounts of text ({} bytes) can be slow.').format(len(btext))
-            if msg:
-                get_boss().confirm(msg + _(' Are you sure?'), partial(self.handle_paste_confirmation, btext), window=self)
+                get_boss().confirm(msg + _(' Are you sure?'), partial(self.handle_large_paste_confirmation, btext), window=self, title=_(
+                'Allow large paste?'))
                 return
         self.paste_text(btext)
 
-    def handle_paste_confirmation(self, btext: bytes, confirmed: bool) -> None:
+    def handle_dangerous_paste_confirmation(self, unsanitized: bytes, sanitized: bytes, choice: str) -> None:
+        if choice == 's':
+            self.paste_text(sanitized)
+        elif choice == 'p':
+            self.paste_text(unsanitized)
+
+    def handle_large_paste_confirmation(self, btext: bytes, confirmed: bool) -> None:
         if confirmed:
             self.paste_text(btext)
 
@@ -1645,14 +1824,13 @@ class Window:
             return None
         return True
 
-    @ac('sc', 'Scroll prompt to the top of the screen, filling screen with empty lines, when in main screen')
+    @ac('sc', 'Scroll prompt to the top of the screen, filling screen with empty lines, when in main screen.'
+        ' To avoid putting the lines above the prompt into the scrollback use scroll_prompt_to_top y')
     def scroll_prompt_to_top(self, clear_scrollback: bool = False) -> Optional[bool]:
         if self.screen.is_main_linebuf():
-            self.screen.scroll_until_cursor_prompt()
-            if clear_scrollback:
-                self.screen.clear_scrollback()
-            elif self.screen.scrolled_by > 0:
-                self.screen.scroll(SCROLL_FULL, False)
+            self.screen.scroll_until_cursor_prompt(not clear_scrollback)
+            if self.screen.scrolled_by > 0:
+                self.scroll_end()
             return None
         return True
 
@@ -1721,3 +1899,36 @@ class Window:
         url = docs_url(which)
         get_boss().open_url(url)
     # }}}
+
+
+def set_pointer_shape(screen: Screen, value: str, os_window_id: int = 0) -> str:
+    op, ret = '=', ''
+    if value and value[0] in '><=?':
+        op = value[0]
+        value = value[1:]
+    if op in '=>':
+        for v in value.split(','):
+            if v or op == '=':
+                screen.change_pointer_shape(op, v)
+        if os_window_id and current_focused_os_window_id() == os_window_id:
+            update_pointer_shape(os_window_id)
+    elif op == '<':
+        screen.change_pointer_shape('<', '')
+        if os_window_id and current_focused_os_window_id() == os_window_id:
+            update_pointer_shape(os_window_id)
+    elif op == '?':
+        ans = []
+        for q in value.split(','):
+            if is_css_pointer_name_valid(q):
+                ans.append('1')
+            else:
+                if q == '__default__':
+                    ans.append(pointer_name_to_css_name(get_options().default_pointer_shape))
+                elif q == '__grabbed__':
+                    ans.append(pointer_name_to_css_name(get_options().pointer_shape_when_grabbed))
+                elif q == '__current__':
+                    ans.append(screen.current_pointer_shape())
+                else:
+                    ans.append('0')
+        ret = ','.join(ans)
+    return ret
