@@ -1,4 +1,4 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import fcntl
@@ -6,6 +6,7 @@ import io
 import os
 import select
 import shlex
+import shutil
 import signal
 import struct
 import sys
@@ -16,7 +17,7 @@ from pty import CHILD, STDIN_FILENO, STDOUT_FILENO, fork
 from unittest import TestCase
 
 from kitty.config import finalize_keys, finalize_mouse_mappings
-from kitty.fast_data_types import Cursor, HistoryBuf, LineBuf, Screen, get_options, parse_bytes, set_options
+from kitty.fast_data_types import Cursor, HistoryBuf, LineBuf, Screen, get_options, set_options
 from kitty.options.parse import merge_result_dicts
 from kitty.options.types import Options, defaults
 from kitty.types import MouseEvent
@@ -24,6 +25,15 @@ from kitty.utils import read_screen_size
 from kitty.window import process_remote_print, process_title_from_child
 
 
+def parse_bytes(screen, data, dump_callback=None):
+    data = memoryview(data)
+    while data:
+        dest = screen.test_create_write_buffer()
+        s = screen.test_commit_write_buffer(data, dest)
+        data = data[s:]
+        screen.test_parse_written_data(dump_callback)
+
+
 class Callbacks:
 
     def __init__(self, pty=None) -> None:
@@ -33,48 +43,48 @@ class Callbacks:
         self.set_pointer_shape = lambda data: None
 
     def write(self, data) -> None:
-        self.wtcbuf += data
+        self.wtcbuf += bytes(data)
 
     def title_changed(self, data, is_base64=False) -> None:
-        self.titlebuf.append(process_title_from_child(data, is_base64))
+        self.titlebuf.append(process_title_from_child(data, is_base64, ''))
 
     def icon_changed(self, data) -> None:
-        self.iconbuf += data
+        self.iconbuf += str(data, 'utf-8')
 
-    def set_dynamic_color(self, code, data) -> None:
+    def set_dynamic_color(self, code, data='') -> None:
         if code == 22:
             self.set_pointer_shape(data)
         else:
-            self.colorbuf += data or ''
+            self.colorbuf += str(data or b'', 'utf-8')
 
-    def set_color_table_color(self, code, data) -> None:
+    def set_color_table_color(self, code, data='') -> None:
         self.ctbuf += ''
 
     def color_profile_popped(self, x) -> None:
         pass
 
+    def cmd_output_marking(self, is_start: bool) -> None:
+        pass
+
     def request_capabilities(self, q) -> None:
         from kitty.terminfo import get_capabilities
         for c in get_capabilities(q, None):
             self.write(c.encode('ascii'))
 
-    def use_utf8(self, on) -> None:
-        self.iutf8 = on
-
-    def desktop_notify(self, osc_code: int, raw_data: str) -> None:
-        self.notifications.append((osc_code, raw_data))
+    def desktop_notify(self, osc_code: int, raw_data: memoryview) -> None:
+        self.notifications.append((osc_code, str(raw_data, 'utf-8')))
 
     def open_url(self, url: str, hyperlink_id: int) -> None:
         self.open_urls.append((url, hyperlink_id))
 
-    def clipboard_control(self, data: str, is_partial: bool = False) -> None:
-        self.cc_buf.append((data, is_partial))
+    def clipboard_control(self, data: memoryview, is_partial: bool = False) -> None:
+        self.cc_buf.append((str(data, 'utf-8'), is_partial))
 
     def clear(self) -> None:
         self.wtcbuf = b''
         self.iconbuf = self.colorbuf = self.ctbuf = ''
         self.titlebuf = []
-        self.iutf8 = True
+        self.printbuf = []
         self.notifications = []
         self.open_urls = []
         self.cc_buf = []
@@ -102,9 +112,13 @@ class Callbacks:
 
     def handle_remote_print(self, msg):
         text = process_remote_print(msg)
-        print(text, file=sys.__stdout__, end='', flush=True)
+        self.printbuf.append(text)
+
+    def handle_remote_cmd(self, msg):
+        pass
 
     def handle_remote_clone(self, msg):
+        msg = str(msg, 'utf-8')
         if not msg:
             if self.current_clone_data:
                 cdata, self.current_clone_data = self.current_clone_data, ''
@@ -166,6 +180,12 @@ class BaseTest(TestCase):
     maxDiff = 2048
     is_ci = os.environ.get('CI') == 'true'
 
+    def rmtree_ignoring_errors(self, tdir):
+        try:
+            shutil.rmtree(tdir)
+        except FileNotFoundError as err:
+            print('Failed to delete the directory:', tdir, 'with error:', err, file=sys.stderr)
+
     def tearDown(self):
         set_options(None)
 
