@@ -1,4 +1,4 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
@@ -13,6 +13,7 @@ from kitty.fast_data_types import (
     LineBuf,
     expand_ansi_c_escapes,
     parse_input_from_terminal,
+    replace_c0_codes_except_nl_space_tab,
     strip_csi,
     truncate_point_for_length,
     wcswidth,
@@ -20,7 +21,7 @@ from kitty.fast_data_types import (
 )
 from kitty.fast_data_types import Cursor as C
 from kitty.rgb import to_color
-from kitty.utils import is_path_in_temp_dir, sanitize_title, sanitize_url_for_dispay_to_user
+from kitty.utils import is_ok_to_read_image_file, is_path_in_temp_dir, sanitize_title, sanitize_url_for_dispay_to_user, shlex_split_with_positions
 
 from . import BaseTest, filled_cursor, filled_history_buf, filled_line_buf
 
@@ -37,6 +38,18 @@ def create_lbuf(*lines):
 
 class TestDataTypes(BaseTest):
 
+
+    def test_replace_c0_codes(self):
+        def t(x: str, expected: str):
+            q = replace_c0_codes_except_nl_space_tab(x)
+            self.ae(expected, q)
+            q = replace_c0_codes_except_nl_space_tab(x.encode('utf-8'))
+            self.ae(expected.encode('utf-8'), q)
+        t('abc', 'abc')
+        t('a\0\x01b\x03\x04\t\rc', 'a\u2400\u2401b\u2403\u2404\t\u240dc')
+        t('a\0\x01😸\x03\x04\t\rc', 'a\u2400\u2401😸\u2403\u2404\t\u240dc')
+        t('a\nb\tc d', 'a\nb\tc d')
+
     def test_to_color(self):
         for x in 'xxx #12 #1234 rgb:a/b'.split():
             self.assertIsNone(to_color(x))
@@ -266,16 +279,17 @@ class TestDataTypes(BaseTest):
         l0 = create('file:///etc/test')
         self.ae(l0.url_start_at(0), 0)
 
-        for trail in '.,\\':
+        for trail in '.,\\}]>':
             lx = create("http://xyz.com" + trail)
             self.ae(lx.url_end_at(0), len(lx) - 2)
-        for trail in ')}]>':
-            lx = create("http://xyz.com" + trail)
-            self.ae(lx.url_end_at(0), len(lx) - 1)
+        for trail in ')':
+            turl = "http://xyz.com" + trail
+            lx = create(turl)
+            self.ae(len(lx) - 1, lx.url_end_at(0), repr(turl))
         l0 = create("ftp://abc/")
         self.ae(l0.url_end_at(0), len(l0) - 1)
         l2 = create("http://-abcd] ")
-        self.ae(l2.url_end_at(0), len(l2) - 2)
+        self.ae(l2.url_end_at(0), len(l2) - 3)
         l3 = create("http://ab.de           ")
         self.ae(l3.url_start_at(4), 0)
         self.ae(l3.url_start_at(5), 0)
@@ -444,6 +458,21 @@ class TestDataTypes(BaseTest):
                 self.assertTrue(is_path_in_temp_dir(os.path.join(prefix, path)))
         for path in ('/home/xy/d.png', '/tmp/../home/x.jpg'):
             self.assertFalse(is_path_in_temp_dir(os.path.join(path)))
+        for path in ('/proc/self/cmdline', os.devnull):
+            if os.path.exists(path):
+                with open(path) as pf:
+                    self.assertFalse(is_ok_to_read_image_file(path, pf.fileno()), path)
+        fifo = os.path.join(tempfile.gettempdir(), 'test-kitty-fifo')
+        os.mkfifo(fifo)
+        fifo_fd = os.open(fifo, os.O_RDONLY | os.O_NONBLOCK)
+        try:
+            self.assertFalse(is_ok_to_read_image_file(fifo, fifo_fd), fifo)
+        finally:
+            os.close(fifo_fd)
+            os.remove(fifo)
+        if os.path.isdir('/dev/shm'):
+            with tempfile.NamedTemporaryFile(dir='/dev/shm') as tf:
+                self.assertTrue(is_ok_to_read_image_file(tf.name, tf.fileno()), fifo)
         self.ae(sanitize_url_for_dispay_to_user(
             'h://a\u0430b.com/El%20Ni%C3%B1o/'), 'h://xn--ab-7kc.com/El Niño/')
 
@@ -597,3 +626,20 @@ class TestDataTypes(BaseTest):
         }.items():
             actual = expand_ansi_c_escapes(src)
             self.ae(expected, actual)
+
+    def test_shlex_split(self):
+        for bad in (
+            'abc\\', '\\', "'abc", "'", '"', 'asd' + '\\',
+        ):
+            with self.assertRaises(ValueError, msg=f'Failed to raise exception for {bad!r}'):
+                tuple(shlex_split_with_positions(bad))
+
+        for q, expected in {
+            '"ab"': ((0, 'ab'),),
+            r'x "ab"y \m': ((0, 'x'), (2, 'aby'), (8, 'm')),
+            r'''x'y"\z'1''': ((0, 'xy"\\z1'),),
+            r'\abc\ d': ((0, 'abc d'),),
+            '': (), '   ': (), ' \tabc\n\t\r ': ((2, 'abc'),),
+        }.items():
+            actual = tuple(shlex_split_with_positions(q))
+            self.ae(expected, actual, f'Failed for text: {q!r}')
