@@ -1,4 +1,4 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
@@ -7,15 +7,13 @@ import tempfile
 import time
 import unittest
 import zlib
-from base64 import standard_b64decode, standard_b64encode
 from contextlib import suppress
 from dataclasses import dataclass
 from io import BytesIO
-from itertools import cycle
 
-from kitty.fast_data_types import load_png_data, parse_bytes, shm_unlink, shm_write, xor_data
+from kitty.fast_data_types import base64_decode, base64_encode, has_avx2, has_sse4_2, load_png_data, shm_unlink, shm_write, test_xor64
 
-from . import BaseTest
+from . import BaseTest, parse_bytes
 
 try:
     from PIL import Image
@@ -28,7 +26,7 @@ def send_command(screen, cmd, payload=b''):
     if payload:
         if isinstance(payload, str):
             payload = payload.encode('utf-8')
-        payload = standard_b64encode(payload).decode('ascii')
+        payload = base64_encode(payload).decode('ascii')
         cmd += ';' + payload
     cmd += '\033\\'
     c = screen.callbacks
@@ -123,13 +121,15 @@ def put_helpers(self, cw, ch, cols=10, lines=5):
         return s, 2 / s.columns, 2 / s.lines
 
     def put_cmd(
-            z=0, num_cols=0, num_lines=0, x_off=0, y_off=0, width=0,
-            height=0, cell_x_off=0, cell_y_off=0, placement_id=0,
-            cursor_movement=0, unicode_placeholder=0
+        z=0, num_cols=0, num_lines=0, x_off=0, y_off=0, width=0, height=0, cell_x_off=0,
+        cell_y_off=0, placement_id=0, cursor_movement=0, unicode_placeholder=0, parent_id=0,
+        parent_placement_id=0, offset_from_parent_x=0, offset_from_parent_y=0,
     ):
-        return 'z=%d,c=%d,r=%d,x=%d,y=%d,w=%d,h=%d,X=%d,Y=%d,p=%d,C=%d,U=%d' % (
-            z, num_cols, num_lines, x_off, y_off, width, height, cell_x_off,
-            cell_y_off, placement_id, cursor_movement, unicode_placeholder
+        return (
+            f'z={z},c={num_cols},r={num_lines},x={x_off},y={y_off},w={width},h={height},'
+            f'X={cell_x_off},Y={cell_y_off},p={placement_id},C={cursor_movement},'
+            f'U={unicode_placeholder},P={parent_id},Q={parent_placement_id},'
+            f'H={offset_from_parent_x},V={offset_from_parent_y}'
         )
 
     def put_image(screen, w, h, **kw):
@@ -183,21 +183,33 @@ def make_send_command(screen):
 class TestGraphics(BaseTest):
 
     def test_xor_data(self):
+        base_data = b'\x01' * 64
+        key = b'\x02' * 64
+        sizes = []
+        if has_sse4_2:
+            sizes.append(2)
+        if has_avx2:
+            sizes.append(3)
+        sizes.append(0)
 
-        def xor(skey, data):
-            ckey = cycle(bytearray(skey))
-            return bytes(bytearray(k ^ d for k, d in zip(ckey, bytearray(data))))
+        def t(key, data, align_offset=0):
+            expected = test_xor64(key, data, 1, 0)
+            for which_function in sizes:
+                actual = test_xor64(key, data, which_function, align_offset)
+                self.ae(expected, actual, f'{align_offset=} {len(data)=}')
 
-        base_data = os.urandom(64)
-        key = os.urandom(len(base_data))
-        for base in (b'', base_data):
+        t(key, b'')
+
+        for base in (b'abc', base_data):
             for extra in range(len(base_data)):
-                data = base + base_data[:extra]
-                self.assertEqual(xor_data(key, data), xor(key, data))
+                for align_offset in range(64):
+                    data = base + base_data[:extra]
+                    t(key, data, align_offset)
 
     def test_disk_cache(self):
         s = self.create_screen()
         dc = s.grman.disk_cache
+        dc.small_hole_threshold = 0
         data = {}
 
         def key_as_bytes(key):
@@ -221,6 +233,13 @@ class TestGraphics(BaseTest):
             for key, val in data.items():
                 self.ae(dc.get(key_as_bytes(key)), val)
 
+        def reset(small_hole_threshold=0):
+            nonlocal dc, data, s
+            s = self.create_screen()
+            dc = s.grman.disk_cache
+            dc.small_hole_threshold = small_hole_threshold
+            data = {}
+
         for i in range(25):
             self.assertIsNone(add(i, f'{i}' * i))
 
@@ -234,6 +253,7 @@ class TestGraphics(BaseTest):
             check_data()
             self.assertRaises(KeyError, dc.get, key_as_bytes(x))
             self.assertEqual(sz, dc.size_on_disk())
+        self.assertEqual(sz, dc.size_on_disk())
         for x in ('xy', 'C'*4, 'B'*6, 'A'*8):
             add(x, x)
             self.assertTrue(dc.wait_for_write())
@@ -287,6 +307,27 @@ class TestGraphics(BaseTest):
         dc.remove_from_ram(clear_predicate)
         self.assertEqual(dc.num_cached_in_ram(), 0)
 
+        reset(512)
+        self.assertIsNone(add(1, '1' * 1024))
+        self.assertIsNone(add(2, '2' * 1024))
+        dc.wait_for_write()
+        sz = dc.size_on_disk()
+        remove(1)
+        self.ae(sz, dc.size_on_disk())
+        self.assertIsNone(add(3, '3' * 800))
+        dc.wait_for_write()
+        self.ae(sz, dc.size_on_disk())
+        self.assertIsNone(add(4, '4' * 100))
+        sz += 100
+        dc.wait_for_write()
+        self.ae(sz, dc.size_on_disk())
+        check_data()
+        remove(4)
+        self.assertIsNone(add(5, '5' * 10))
+        sz += 10
+        dc.wait_for_write()
+        self.ae(sz, dc.size_on_disk())
+
     def test_suppressing_gr_command_responses(self):
         s, g, pl, sl = load_helpers(self)
         self.ae(pl('abcd', s=10, v=10, q=1), 'ENODATA:Insufficient image data: 4 < 400')
@@ -340,20 +381,27 @@ class TestGraphics(BaseTest):
         img = g.image_for_client_id(1)
         self.ae(img['data'], b'abcdefghijklmnop')
 
+        random_data = byte_block(32 * 1024)
+        sl(
+            random_data,
+            s=1024,
+            v=8,
+            expecting_data=random_data
+        )
+
         # Test compression
-        random_data = byte_block(3 * 1024)
         compressed_random_data = zlib.compress(random_data)
         sl(
             compressed_random_data,
-            s=24,
-            v=32,
+            s=1024,
+            v=8,
             o='z',
             expecting_data=random_data
         )
 
         # Test chunked + compressed
         b = len(compressed_random_data) // 2
-        self.assertIsNone(pl(compressed_random_data[:b], s=24, v=32, o='z', m=1))
+        self.assertIsNone(pl(compressed_random_data[:b], s=1024, v=8, o='z', m=1))
         self.ae(pl(compressed_random_data[b:], m=0), 'OK')
         img = g.image_for_client_id(1)
         self.ae(img['data'], random_data)
@@ -362,10 +410,10 @@ class TestGraphics(BaseTest):
         def load_temp(prefix='tty-graphics-protocol-'):
             f = tempfile.NamedTemporaryFile(prefix=prefix)
             f.write(random_data), f.flush()
-            sl(f.name, s=24, v=32, t='f', expecting_data=random_data)
+            sl(f.name, s=1024, v=8, t='f', expecting_data=random_data)
             self.assertTrue(os.path.exists(f.name))
             f.seek(0), f.truncate(), f.write(compressed_random_data), f.flush()
-            sl(f.name, s=24, v=32, t='t', o='z', expecting_data=random_data)
+            sl(f.name, s=1024, v=8, t='t', o='z', expecting_data=random_data)
             return f
 
         f = load_temp()
@@ -379,7 +427,7 @@ class TestGraphics(BaseTest):
         # Test loading from POSIX SHM
         name = '/kitty-test-shm'
         shm_write(name, random_data)
-        sl(name, s=24, v=32, t='s', expecting_data=random_data)
+        sl(name, s=1024, v=8, t='s', expecting_data=random_data)
         self.assertRaises(
             FileNotFoundError, shm_unlink, name
         )  # check that file was deleted
@@ -419,7 +467,7 @@ class TestGraphics(BaseTest):
 
     def test_load_png_simple(self):
         # 1x1 transparent PNG
-        png_data = standard_b64decode('iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNk+P+/HgAFhAJ/wlseKgAAAABJRU5ErkJggg==')
+        png_data = base64_decode('iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNk+P+/HgAFhAJ/wlseKgAAAABJRU5ErkJggg==')
         expected = b'\x00\xff\xff\x7f'
         self.ae(load_png_data(png_data), (expected, 1, 1))
         s, g, pl, sl = load_helpers(self)
@@ -540,6 +588,105 @@ class TestGraphics(BaseTest):
         self.ae(put_image(s, 8, 16, id=2, z=-1)[1], 'OK')
         self.ae(group_counts(), (2, 1, 1, 2, 1))
 
+    def test_image_parents(self):
+        cw, ch = 10, 20
+        iw, ih = 10, 20
+        s, dx, dy, put_image, put_ref, layers, rect_eq = put_helpers(self, cw, ch)
+
+        def positions():
+            ans = {}
+            def x(x):
+                return round(((x + 1)/2) * s.columns)
+            def y(y):
+                return int(((-y + 1)/2) * s.lines)
+
+            for i in layers(s):
+                d = i['dest_rect']
+                ans[(i['image_id'], i['ref_id'])] = {'x': x(d['left']), 'y': y(d['top'])}
+            return ans
+
+        def p(x, y=0):
+            return {'x':x, 'y': y}
+
+        self.ae(put_image(s, iw, ih, id=1)[1], 'OK')
+        self.ae(put_ref(s, id=1, placement_id=1), (1, ('OK', 'i=1,p=1')))
+        pos = {(1, 1): p(0), (1, 2): p(1)}
+        self.ae(positions(), pos)
+        # check that adding a reference to a non-existent parent fails
+        self.ae(put_ref(s, id=1, placement_id=33, parent_id=1, parent_placement_id=2), (1, ('ENOPARENT', 'i=1,p=33')))
+        self.ae(put_ref(s, id=1, placement_id=33, parent_id=33), (1, ('ENOPARENT', 'i=1,p=33')))
+        # check that we cannot add a reference that is its own parent
+        self.ae(put_ref(s, id=1, placement_id=1, parent_id=1, parent_placement_id=1), (1, ('EINVAL', 'i=1,p=1')))
+
+        self.ae(put_image(s, iw, ih, id=2)[1], 'OK')
+        pos[(2,1)] = p(2)
+        self.ae(positions(), pos)
+        # Add two children to the first placement of img2
+        before = s.cursor.x, s.cursor.y
+        self.ae(put_ref(s, id=1, placement_id=2, parent_id=2, offset_from_parent_y=3), (1, ('OK', 'i=1,p=2')))
+        self.ae(before, (s.cursor.x, s.cursor.y), 'Cursor must not move for child image')
+        pos[(1,3)] = p(2, 3)
+        self.ae(positions(), pos)
+        self.ae(put_ref(s, id=2, placement_id=3, parent_id=2, offset_from_parent_y=4), (2, ('OK', 'i=2,p=3')))
+        pos[(2,2)] = p(2, 4)
+        self.ae(positions(), pos)
+        # Add a grand child to the second child of img2
+        self.ae(put_ref(s, id=2, placement_id=4, parent_id=2, parent_placement_id=3, offset_from_parent_x=-1), (2, ('OK', 'i=2,p=4')))
+        pos[(2,3)] = p(pos[(2,2)]['x']-1, pos[(2,2)]['y'])
+        self.ae(positions(), pos)
+        # Check that creating a cycle is prevented
+        self.ae(put_ref(s, id=2, placement_id=3, parent_id=2, parent_placement_id=4), (2, ('ECYCLE', 'i=2,p=3')))
+        self.ae(positions(), pos)
+        # Check that depth is limited
+        for i in range(5, 12):
+            q = put_ref(s, id=2, placement_id=i, parent_id=2, parent_placement_id=i-1, offset_from_parent_x=-1)[1][0]
+            if q == 'ETOODEEP':
+                break
+            self.ae(q, 'OK')
+        else:
+            self.assertTrue(False, 'Failed to limit reference chain depth')
+        # Check that deleting a parent removes all descendants
+        send_command(s, 'a=d,d=i,i=2,p=3')
+        pos.pop((2,3)), pos.pop((2,2))
+        self.ae(positions(), pos)
+        # Check that deleting a parent deletes all descendants and also removes
+        # images with no remaining placements
+        self.ae(put_ref(s, id=2, placement_id=3, parent_id=2, offset_from_parent_y=4), (2, ('OK', 'i=2,p=3')))
+        pos[(2,11)] = p(2, 4)
+        self.ae(positions(), pos)
+        self.ae(put_image(s, iw, ih, id=3, placement_id=97, parent_id=2, parent_placement_id=3)[1], 'OK')
+        pos[(3,1)] = p(2, 4)
+        self.ae(positions(), pos)
+        send_command(s, 'a=d,d=i,i=2')
+        pos.pop((3,1)), pos.pop((2,11)), pos.pop((2,1)), pos.pop((1,3))
+        self.ae(positions(), pos)
+        # Check that virtual placements that try to be relative are rejected
+        self.ae(put_ref(s, id=1, placement_id=11, parent_id=1, unicode_placeholder=1), (1, ('EINVAL', 'i=1,p=11')))
+        # Check creation of children of a unicode placeholder based image
+        s, dx, dy, put_image, put_ref, layers, rect_eq = put_helpers(self, cw, ch)
+        put_image(s, 20, 20, num_cols=4, num_lines=2, unicode_placeholder=1, id=42)
+        s.update_only_line_graphics_data()
+        self.assertFalse(positions())  # the reference is virtual
+        self.ae(put_ref(s, id=42, placement_id=11, parent_id=42, offset_from_parent_y=2, offset_from_parent_x=1), (42, ('OK', 'i=42,p=11')))
+        self.assertFalse(positions())  # the reference is virtual without any cell images so the child is invisible
+        s.apply_sgr("38;5;42")
+        # These two characters will become one 2x1 ref.
+        s.cursor.x = s.cursor.y = 1
+        s.draw("\U0010EEEE\u0305\u0305\U0010EEEE\u0305\u030D")
+        s.cursor.x = s.cursor.y = 0
+        s.draw("\U0010EEEE\u0305\u0305\U0010EEEE\u0305\u030D")
+        s.update_only_line_graphics_data()
+        pos = {(1, 2): p(1, 2), (1, 3): p(0), (1, 4): p(1)}
+        self.ae(positions(), pos)
+        s.cursor.x = s.cursor.y = 0
+        s.erase_in_display(0, False)
+        s.update_only_line_graphics_data()
+        self.assertFalse(positions())  # the reference is virtual without any cell images so the child is invisible
+        s.cursor.x = s.cursor.y = 2
+        s.draw("\U0010EEEE\u0305\u0305\U0010EEEE\u0305\u030D")
+        s.update_only_line_graphics_data()
+        self.ae(positions(), {(1, 5): {'x': 2, 'y': 2}, (1, 2): {'x': 3, 'y': 4}})
+
     def test_unicode_placeholders(self):
         # This test tests basic image placement using using unicode placeholders
         cw, ch = 10, 20
@@ -879,6 +1026,13 @@ class TestGraphics(BaseTest):
         put_image(s, cw, ch, z=9)
         delete('Z', z=9)
         self.ae(s.grman.image_count, 0)
+        put_image(s, cw, ch, id=1)
+        put_image(s, cw, ch, id=2)
+        put_image(s, cw, ch, id=3)
+        delete('R', y=2)
+        self.ae(s.grman.image_count, 1)
+        delete('R', x=3, y=3)
+        self.ae(s.grman.image_count, 0)
         self.assertEqual(s.grman.disk_cache.total_size, 0)
 
         # test put + delete + put
