@@ -12,7 +12,7 @@ from contextlib import contextmanager
 from functools import lru_cache, partial
 
 from kitty.bash import decode_ansi_c_quoted_string
-from kitty.constants import kitty_base_dir, shell_integration_dir, terminfo_dir
+from kitty.constants import kitten_exe, kitty_base_dir, shell_integration_dir, terminfo_dir
 from kitty.fast_data_types import CURSOR_BEAM, CURSOR_BLOCK, CURSOR_UNDERLINE
 from kitty.shell_integration import setup_bash_env, setup_fish_env, setup_zsh_env
 
@@ -41,6 +41,7 @@ def basic_shell_env(home_dir):
         'KITTY_INSTALLATION_DIR': kitty_base_dir,
         'BASH_SILENCE_DEPRECATION_WARNING': '1',
         'PYTHONDONTWRITEBYTECODE': '1',
+        'WEZTERM_SHELL_SKIP_ALL': '1',  # dont fail if WezTerm's system wide, default on (why?) shell integration is installed
     }
     for x in ('USER', 'LANG'):
         if os.environ.get(x):
@@ -48,7 +49,7 @@ def basic_shell_env(home_dir):
     return ans
 
 
-def safe_env_for_running_shell(argv, home_dir, rc='', shell='zsh'):
+def safe_env_for_running_shell(argv, home_dir, rc='', shell='zsh', with_kitten=False):
     ans = basic_shell_env(home_dir)
     if shell == 'zsh':
         argv.insert(1, '--noglobalrcs')
@@ -64,10 +65,14 @@ def safe_env_for_running_shell(argv, home_dir, rc='', shell='zsh'):
             print(rc + '\n', file=f)
         setup_fish_env(ans, argv)
     elif shell == 'bash':
-        setup_bash_env(ans, argv)
-        ans['KITTY_BASH_INJECT'] += ' posix'
-        ans['KITTY_BASH_POSIX_ENV'] = os.path.join(home_dir, '.bashrc')
-        with open(ans['KITTY_BASH_POSIX_ENV'], 'w') as f:
+        bashrc = os.path.join(home_dir, '.bashrc')
+        if with_kitten:
+            ans['KITTY_RUNNING_BASH_INTEGRATION_TEST'] = bashrc
+        else:
+            setup_bash_env(ans, argv)
+            ans['KITTY_BASH_INJECT'] += ' posix'
+            ans['KITTY_BASH_POSIX_ENV'] = bashrc
+        with open(bashrc, 'w') as f:
             # ensure LINES and COLUMNS are kept up to date
             print('shopt -s checkwinsize', file=f)
             if rc:
@@ -77,13 +82,18 @@ def safe_env_for_running_shell(argv, home_dir, rc='', shell='zsh'):
 
 class ShellIntegration(BaseTest):
 
+    with_kitten = False
+
     @contextmanager
     def run_shell(self, shell='zsh', rc='', cmd='', setup_env=None):
         home_dir = self.home_dir = os.path.realpath(tempfile.mkdtemp())
         cmd = cmd or shell
         cmd = shlex.split(cmd.format(**locals()))
-        env = (setup_env or safe_env_for_running_shell)(cmd, home_dir, rc=rc, shell=shell)
+        env = (setup_env or safe_env_for_running_shell)(cmd, home_dir, rc=rc, shell=shell, with_kitten=self.with_kitten)
+        env['KITTY_RUNNING_SHELL_INTEGRATION_TEST'] = '1'
         try:
+            if self.with_kitten:
+                cmd = [kitten_exe(), 'run-shell', '--shell', shlex.join(cmd)]
             pty = self.create_pty(cmd, cwd=home_dir, env=env)
             i = 10
             while i > 0 and not pty.screen_contents().strip():
@@ -143,12 +153,12 @@ RPS1="{rps1}"
             pty.callbacks.clear()
             pty.send_cmd_to_child('printf "%s\x16\a%s" "a" "b"')
             pty.wait_till(lambda: 'ab' in pty.screen_contents())
-            self.assertTrue(pty.screen.last_reported_cwd.endswith(self.home_dir))
+            self.assertTrue(pty.screen.last_reported_cwd.decode().endswith(self.home_dir))
             self.assertIn('%s^G%s', pty.screen_contents())
             q = os.path.join(self.home_dir, 'testing-cwd-notification-🐱')
             os.mkdir(q)
             pty.send_cmd_to_child(f'cd {q}')
-            pty.wait_till(lambda: pty.screen.last_reported_cwd.endswith(q))
+            pty.wait_till(lambda: pty.screen.last_reported_cwd.decode().endswith(q))
         with self.run_shell(rc=f'''PS1="{ps1}"\nexport ES="a\n b c\nd"''') as pty:
             pty.callbacks.clear()
             pty.send_cmd_to_child('clone-in-kitty')
@@ -179,13 +189,13 @@ function _set_status_prompt; function fish_prompt; echo -n "$pipestatus $status
             pty.wait_till(lambda: 'XDD_OK' in pty.screen_contents())
 
             # CWD reporting
-            self.assertTrue(pty.screen.last_reported_cwd.endswith(self.home_dir))
+            self.assertTrue(pty.screen.last_reported_cwd.decode().endswith(self.home_dir))
             q = os.path.join(self.home_dir, 'testing-cwd-notification-🐱')
             os.mkdir(q)
             pty.send_cmd_to_child(f'cd {q}')
-            pty.wait_till(lambda: pty.screen.last_reported_cwd.endswith(q))
+            pty.wait_till(lambda: pty.screen.last_reported_cwd.decode().endswith(q))
             pty.send_cmd_to_child('cd -')
-            pty.wait_till(lambda: pty.screen.last_reported_cwd.endswith(self.home_dir))
+            pty.wait_till(lambda: pty.screen.last_reported_cwd.decode().endswith(self.home_dir))
 
             # completion and prompt marking
             pty.wait_till(lambda: 'cd -' not in pty.screen_contents().splitlines()[-1])
@@ -295,13 +305,13 @@ PS1="{ps1}"
             pty.send_cmd_to_child('printf "%s\x16\a%s" "a" "b"')
             pty.wait_till(lambda: pty.screen_contents().count(ps1) == 2)
             self.ae(pty.screen_contents(), f'{ps1}printf "%s^G%s" "a" "b"\nab{ps1}')
-            self.assertTrue(pty.screen.last_reported_cwd.endswith(self.home_dir))
+            self.assertTrue(pty.screen.last_reported_cwd.decode().endswith(self.home_dir))
             pty.send_cmd_to_child('echo $HISTFILE')
-            pty.wait_till(lambda: '.bash_history' in pty.screen_contents())
+            pty.wait_till(lambda: '.bash_history' in pty.screen_contents().replace('\n', ''))
             q = os.path.join(self.home_dir, 'testing-cwd-notification-🐱')
             os.mkdir(q)
             pty.send_cmd_to_child(f'cd {q}')
-            pty.wait_till(lambda: pty.screen.last_reported_cwd.endswith(q))
+            pty.wait_till(lambda: pty.screen.last_reported_cwd.decode().endswith(q))
 
         for ps1 in ('line1\\nline\\2\\prompt> ', 'line1\nprompt> ', 'line1\\nprompt> ',):
             with self.subTest(ps1=ps1), self.run_shell(
@@ -325,9 +335,10 @@ PS1="{ps1}"
 
         # test startup file sourcing
 
-        def setup_env(excluded, argv, home_dir, rc='', shell='bash'):
+        def setup_env(excluded, argv, home_dir, rc='', shell='bash', with_kitten=self.with_kitten):
             ans = basic_shell_env(home_dir)
-            setup_bash_env(ans, argv)
+            if not with_kitten:
+                setup_bash_env(ans, argv)
             for x in {'profile', 'bash.bashrc', '.bash_profile', '.bash_login', '.profile', '.bashrc', 'rcfile'} - excluded:
                 with open(os.path.join(home_dir, x), 'w') as f:
                     if x == '.bashrc' and rc:
@@ -381,3 +392,7 @@ PS1="{ps1}"
         }.items():
             q = q + "'"
             self.ae(decode_ansi_c_quoted_string(q, 0)[0], e, f'Failed to decode: {q!r}')
+
+
+class ShellIntegrationWithKitten(ShellIntegration):
+    with_kitten = True
