@@ -1,4 +1,4 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>
 
 import argparse
@@ -36,8 +36,8 @@ ap = re.search(r"^appname: str\s+=\s+'([^']+)'", raw, flags=re.MULTILINE)
 if ap is not None:
     appname = ap.group(1)
 
-ALL_ACTIONS = 'man html build tag sdist upload website'.split()
-NIGHTLY_ACTIONS = 'man html build sdist upload_nightly'.split()
+ALL_ACTIONS = 'local_build man html build tag sdist upload website'.split()
+NIGHTLY_ACTIONS = 'local_build man html build sdist upload_nightly'.split()
 
 
 def echo_cmd(cmd: Iterable[str]) -> None:
@@ -61,6 +61,10 @@ def call(*cmd: str, cwd: Optional[str] = None, echo: bool = False) -> None:
         raise SystemExit(ret)
 
 
+def run_local_build(args: Any) -> None:
+    call('make debug')
+
+
 def run_build(args: Any) -> None:
     import runpy
 
@@ -366,7 +370,7 @@ class GitHub:  # {{{
         if self.is_nightly:
             for fname in tuple(assets_by_fname):
                 self.info(f'Deleting {fname} from GitHub with id: {assets_by_fname[fname]["id"]}')
-                delete_asset(assets_by_fname.pop(fname), allow_not_found=False)
+                delete_asset(assets_by_fname.pop(fname))
         for path, desc in self.files.items():
             self.info('')
             upload_with_retries(path, desc)
