@@ -23,7 +23,6 @@ const (
 const (
 	SAVE_CURSOR                   = "\0337"
 	RESTORE_CURSOR                = "\0338"
-	S7C1T                         = "\033 F"
 	SAVE_PRIVATE_MODE_VALUES      = "\033[?s"
 	RESTORE_PRIVATE_MODE_VALUES   = "\033[?r"
 	SAVE_COLORS                   = "\033[#P"
@@ -95,7 +94,7 @@ const (
 )
 
 type TerminalStateOptions struct {
-	alternate_screen, restore_colors bool
+	Alternate_screen, restore_colors bool
 	mouse_tracking                   MouseTracking
 	kitty_keyboard_mode              KeyboardStateBits
 }
@@ -115,8 +114,7 @@ func reset_modes(sb *strings.Builder, modes ...Mode) {
 func (self *TerminalStateOptions) SetStateEscapeCodes() string {
 	var sb strings.Builder
 	sb.Grow(256)
-	sb.WriteString(S7C1T)
-	if self.alternate_screen {
+	if self.Alternate_screen {
 		sb.WriteString(SAVE_CURSOR)
 	}
 	sb.WriteString(SAVE_PRIVATE_MODE_VALUES)
@@ -128,7 +126,7 @@ func (self *TerminalStateOptions) SetStateEscapeCodes() string {
 		IRM, DECKM, DECSCNM, BRACKETED_PASTE, FOCUS_TRACKING,
 		MOUSE_BUTTON_TRACKING, MOUSE_MOTION_TRACKING, MOUSE_MOVE_TRACKING, MOUSE_UTF8_MODE, MOUSE_SGR_MODE)
 	set_modes(&sb, DECARM, DECAWM, DECTCEM)
-	if self.alternate_screen {
+	if self.Alternate_screen {
 		set_modes(&sb, ALTERNATE_SCREEN)
 		sb.WriteString(CLEAR_SCREEN)
 	}
@@ -155,16 +153,16 @@ func (self *TerminalStateOptions) ResetStateEscapeCodes() string {
 	var sb strings.Builder
 	sb.Grow(64)
 	sb.WriteString("\033[<u")
-	if self.alternate_screen {
+	if self.Alternate_screen {
 		sb.WriteString(ALTERNATE_SCREEN.EscapeCodeToReset())
 	} else {
 		sb.WriteString(SAVE_CURSOR)
 	}
 	sb.WriteString(RESTORE_PRIVATE_MODE_VALUES)
 	if self.restore_colors {
-		sb.WriteString(RESTORE_CURSOR)
+		sb.WriteString(RESTORE_COLORS)
 	}
-	sb.WriteString(RESTORE_COLORS)
+	sb.WriteString(RESTORE_CURSOR)
 	return sb.String()
 }
 
